from .worker import Worker

from ..mesh.mesh_attributes import Attribute
from ..mesh.datatypes import *
from ..mesh.mesh_data import RawMeshData
from .. import geometry
from ..attributes.misc_faces import face_normals
from ..attributes.misc_corners import corner_angles
from math import pi

class FeatureEdgeDetector(Worker):
    """
    Worker used to detect features on a surface mesh. Feature edges are of three types:
    
    - Marked edges from the input file (reading the "hard_edges" attribute on edges)
    
    - Boundary edges
    
    - Edges such that their dihedral angle is large (crease edges)

    The `FeatureEdgeDetector` object ican be given as a parameter in parametrization or frame field algorithms that need feature edge alignment.
    """

    def __init__(self, only_border : bool = False, flag_corners=True, corner_order:int = 4, compute_feature_graph=True, verbose=True):
        """
        Parameters:
            only_border (bool, optional): If set to True, will only consider border edges as features. Defaults to False.
            flag_corners (bool, optional): If set to True, will also compute a goal angle defect (multiple of pi/2) of each detected vertices. Defaults to True.
            corner_order (int, optional): For corner detection, considers corners of angle defect 2pi/corner_order. Defaults to 4 (corners of pi/2).
            compute_feature_graph (bool, optional): whether to compute a Polyline object representing the feature graph. For debug and visualization purposes. Defaults to True.
            verbose (bool, optional): Verbose mode. Defaults to True.
        
        Attributes:
            feature_vertices (set): indices of the vertices that are adjacent to at least one feature edge
            feature_edges (set): indices of feature edges
            feature_degree (Attribute): number of feature edges each vertex is adjacent to
            local_feat_edges (dict): features edges in the neighborhood of each vertex (in the order of `mesh.connectivity.vertex_to_edges`)
            corners (Attribute): the order of each detected vertex corners
        """
        self._feature_graph : PolyLine = None
        self.only_border: bool = only_border # whether to ignore every feature edge that is not a boundary
        self.flag_corners: bool = flag_corners # whether to also flag the angle value of each feature vertex
        self.corner_order: int = corner_order
        self.compute_feature_graph: bool = compute_feature_graph
        
        self.border_cycles : list = None # result of 'M.processing.extract_border_cycle_all'

        self.fnormals : Attribute = None # face normals

        # Following data structures will be filled
        self.feature_vertices : set = None # indexes of feature vertices
        self.feature_edges : set = None # indexes of feature edges
        self.feature_degrees : Attribute = None # degree of each feature vertex in the feature graph
        self.local_feat_edges : dict = None # vertex -> list of local indices of edges that are feature edges
        self.corners : Attribute = None # index of each detected corners (as an int k such that defect is close to k*pi/2)

        super().__init__("FeatureDetector", verbose)

    def clear(self):
        """Clears the data structures"""
        self.feature_vertices = set()
        self.feature_edges = set()
        self.feature_degrees = Attribute(int)
        self.local_feat_edges = dict() # vertex -> list of local indices of edges that are feature edges



##### Feature Graph #####

    @property
    def feature_graph(self) -> PolyLine:
        """the feature edges as a polyline object"""
        if self._feature_graph is None:
            self._compute_feature_graph()
        return self._feature_graph

    def _compute_feature_graph(self, mesh: SurfaceMesh):
        """Computes self._feature_mesh"""
        self._feature_graph = RawMeshData()
        fvert = list(self.feature_vertices)
        fvert = dict([(fvert[i],i) for i in range(len(self.feature_vertices))])
        degree_attr = self._feature_graph.vertices.create_attribute("degree", int)
        self._feature_graph.vertices += [mesh.vertices[k] for k in fvert.keys()]
        for v in self.feature_vertices:
            degree_attr[fvert[v]] = self.feature_degrees[v]
        for e in self.feature_edges:
            A, B = mesh.edges[e]
            self._feature_graph.edges.append((fvert[A],fvert[B]))
        self._feature_graph = PolyLine(self._feature_graph)
 
##### Feature detection subfunctions #####

    def _add_border_to_features(self, mesh : SurfaceMesh, feature_attr : Attribute) -> Attribute:
        if len(mesh.boundary_edges)==0 : return feature_attr # no border
        for e in mesh.boundary_edges:
            feature_attr[e] = True
        return feature_attr

    def _add_hard_edges_to_features(self, mesh : SurfaceMesh, feature_attr : Attribute) -> Attribute:
        """The mesh may already define a set of "hard" edges, for instance if it was imported from a file where edges were specified.
        These edges should be registered as features, but we filter them if their curvature is too small.

        Parameters:
            feature_attr (Attribute): the feature flag to fill in

        Returns:
            Attribute: the modified feature flag
        """
        if self.only_border : return feature_attr
        DOT_THRESHOLD = 0.2
        if mesh.edges.has_attribute("hard_edges"):
            for e in mesh.edges.get_attribute("hard_edges"):
                A,B = mesh.edges[e]
                T1,T2 = mesh.connectivity.edge_to_faces(A,B)
                if T1 is None or T2 is None : continue
                N1,N2 = self.fnormals[T1], self.fnormals[T2]
                # we filter hard edge also depending on their angles. if marked but flat, the edge is ignored
                if geometry.dot(N1,N2) < 1 - DOT_THRESHOLD and not mesh.is_edge_on_border(*mesh.edges[e]):
                    feature_attr[e] = True
        return feature_attr

    def _add_sharp_angles_to_features(self, mesh : SurfaceMesh, feature_attr : Attribute) -> Attribute:
        """Flags as features all edges which dihedral angle is smaller than pi/2 or greater than 3*pi/2

        Parameters:
            feature_attr (Attribute): the feature flag to fill in

        Returns:
            Attribute: the modified feature flag
        """
        if self.only_border : return feature_attr
        DOT_THRESHOLD = 0.5
        for e, (A,B) in enumerate(mesh.edges):
            T1,T2 = mesh.connectivity.edge_to_faces(A,B)
            if T1 is None or T2 is None : continue
            N1,N2 = self.fnormals[T1], self.fnormals[T2]
            if geometry.dot(N1,N2) < DOT_THRESHOLD:
                feature_attr[e] = True
        return feature_attr
    
    def _flag_corners(self, mesh : SurfaceMesh):
        if mesh.vertices.has_attribute("corners"):
            self.corners = mesh.vertices.get_attribute("corners")
        else:
            self.corners = mesh.vertices.create_attribute("corners", int)
        angles = corner_angles(mesh, persistent=False)
        for v in self.feature_vertices:
            angle_v = 0.
            for T in mesh.connectivity.vertex_to_faces(v):
                c = mesh.connectivity.vertex_to_corner_in_face(v,T)
                angle_v += angles[c]
            if abs(angle_v) <  2*pi/self.corner_order:
                self.corners[v] = 1 if angle_v>=0 else -1
            else:
                self.corners[v] = round(angle_v * self.corner_order / ( 2 * pi))

##### Main detect function #####

    @allowed_mesh_types(SurfaceMesh)
    def run(self, mesh : SurfaceMesh):
        self.clear()
        self.log("Initializing attributes")
        if mesh.faces.has_attribute("normals"):
            self.fnormals = mesh.faces.get_attribute("normals")
        else:
            self.fnormals = face_normals(mesh, persistent=False)

        # create the feature attribute
        if mesh.vertices.has_attribute("feature"):
            feat_v = mesh.vertices.get_attribute("feature")
            feat_v.clear()
        else:
            feat_v = mesh.vertices.create_attribute("feature", bool)
        
        if mesh.edges.has_attribute("feature"):
            feature = mesh.edges.get_attribute("feature")
            feature.clear()
        else:
            feature = mesh.edges.create_attribute("feature", bool)

        self.log("Detect Features")
        feature = self._add_hard_edges_to_features(mesh, feature)
        feature = self._add_sharp_angles_to_features(mesh, feature)
        feature = self._add_border_to_features(mesh, feature)

        self.log("Build set containers")
        for e in feature:
            A,B = mesh.edges[e]
            self.feature_edges.add(e)
            self.feature_vertices.add(A)
            self.feature_vertices.add(B)
        
        self.log("Build local feature edges indices")
        for v in self.feature_vertices:
            self.local_feat_edges[v] = []
            for i, ev in enumerate(mesh.connectivity.vertex_to_edges(v)):
                if feature[ev] :
                    self.local_feat_edges[v].append(i)

        for e in self.feature_edges:
            A,B = mesh.edges[e]
            self.feature_degrees[A] += 1
            self.feature_degrees[B] += 1

        if self.flag_corners:
            self.log("Flag corners")
            self._flag_corners(mesh)

        if self.compute_feature_graph:
            self.log("Compute feature graph")
            self._compute_feature_graph(mesh)

        self.log(f" # Feature edges: {len(self.feature_edges)}")

        for v in self.feature_vertices: feat_v[v] = True
        self.log(f" # Feature vertices: {len(self.feature_vertices)}")


    def detect(self, mesh : SurfaceMesh):
        """Runs the detection on a provided mesh.

        Args:
            mesh (SurfaceMesh): the input mesh
        """
        self.run(mesh)


from ..mesh_data import RawMeshData
from ..mesh_attributes import Attribute
from .base import Mesh
from .linear import PolyLine
from ... import utils
from ... import config

class SurfaceMesh(Mesh):
    """A data structure for representing polygonal surfaces.
    
    Attributes:
        vertices (DataContainer): the container for all vertices
        edges (DataContainer): the container for all edges
        faces (DataContainer): the container for all faces
        face_corners (DataContainer): the container for all corner of faces

        boundary_edges (list): list of all edge indices on the boundary
        interior_edges (list): list of all interior edge indices (all edges \\ boundary_edges)
        boundary_vertices (list): list of all vertex indices on the boundary
        interior_vertices (list): list of all interior verticex indices (all vertices \\ boundary_vertices)

        connectivity (_SurfaceConnectivity): the connectivity utility class
    """

    def __init__(self, data : RawMeshData = None):
        Mesh.__init__(self, 2, data)

        self.connectivity = SurfaceMesh._Connectivity(self)

        # Boundary data
        self._boundary_edges : list = None
        self._interior_edges : list = None

        self._is_vertex_on_border : Attribute = None
        self._boundary_vertices : list = None
        self._interior_vertices : list = None

        self._is_triangular : bool = None
        self._is_quad : bool = None

    def __str__(self):
        out = "mouette.mesh.SurfaceMesh object\n"
        out += "| {} vertices\n".format(len(self.vertices))
        out += "| {} edges\n".format(len(self.edges))
        out += "| {} faces\n".format(len(self.faces))
        return out

    @property
    def id_vertices(self):
        """
        Shortcut for `range(len(self.vertices))`
        """
        return range(len(self.vertices))

    @property
    def id_edges(self):
        """
        Shortcut for `range(len(self.edges))`
        """
        return range(len(self.edges))

    @property
    def id_faces(self):
        """
        Shortcut for `range(len(self.faces))`
        """
        return range(len(self.faces))

    @property
    def id_corners(self):
        """
        Shortcut for `range(len(self.face_corners))`
        """
        return range(len(self.face_corners))

    def ith_vertex_of_face(self, fid: int, i: int) -> int:
        """
        helper function to get the i-th vertex of a face, i.e. `self.faces[fid][i]`

        Args:
            fid (int): face id
            i (int): vertex id in face. Should be 0 <= vid < len(face)

        Returns:
            int: the id of the i-th vertex in face `fid` (`self.faces[fid][i]`)
        """
        return self.faces[fid][i]

    def pt_of_face(self, fid:int):
        """
        point coordinates of vertices of face `fid`

        Args:
            fid (int): face id

        Returns:
            Iterable: iterator of Vec objects representing point coordinates of vertices
        """
        return (self.vertices[_v] for _v in self.faces[fid])

    def _compute_mesh_type(self):
        self._is_triangular = True
        self._is_quad = True
        for f in self.faces:
            self._is_triangular = self._is_triangular and len(f)==3
            self._is_quad = self._is_quad and len(f)==4

    def is_triangular(self) -> bool:
        """
        Checks if the mesh is a triangulation

        Returns:
            bool: True if the mesh is triangular (all faces are triangles)
        """
        if self._is_triangular is None:
            self._compute_mesh_type()
        return self._is_triangular

    def is_quad(self) -> bool:
        """
        Checks if the mesh is a quadrangulation

        Returns:
            bool: True if the mesh is quadrangular (all faces are quad)
        """
        if self._is_quad is None:
            self._compute_mesh_type()
        return self._is_quad

    def clear_boundary_data(self):
        """
        Clear all boundary data.
        Next call to a boundary/interior container or method will recompute everything
        """
        self._boundary_edges : list = None
        self._interior_edges : list = None

        self._is_vertex_on_border : Attribute = None 
        self._boundary_vertices : set = None
        self._interior_vertices : set = None

    ##### Accessors for boundary and exterior ######

    def _compute_interior_boundary_edges(self):
        self._interior_edges = []
        self._boundary_edges = []
        for e,(u,v) in enumerate(self.edges):
            if self.is_edge_on_border(u,v):
                self._boundary_edges.append(e)
            else:
                self._interior_edges.append(e)
    
    def _compute_interior_boundary_vertices(self):
        self._boundary_vertices = set()
        self._is_vertex_on_border = self.vertices.create_attribute("border", bool)
        for e in self.boundary_edges:
            a,b = self.edges[e]
            self._is_vertex_on_border[a] = True
            self._is_vertex_on_border[b] = True
            self._boundary_vertices.add(a)
            self._boundary_vertices.add(b)
        self._boundary_vertices = list(self.boundary_vertices)
        
        self._interior_vertices = []
        for x in self.id_vertices:
            if not self._is_vertex_on_border[x]:
                self._interior_vertices.append(x)

    def is_edge_on_border(self, u:int, v:int) -> bool:
        """
        whether edge (u,v) is a boundary edge or not

        Args:
            u (int): vertex id
            v (int): vertex id

        Returns:
            bool: whether edge (u,v) is a boundary edge or not. Returns False if (u,v) is not a valid edge.
        """
        if self.connectivity.edge_id(u,v) is None: return False
        return self.connectivity.direct_face(u,v) is None or self.connectivity.direct_face(v,u) is None

    def is_vertex_on_border(self, u:int) -> bool:
        """
        whether vertex `u` is a boundary vertex or not.

        Args:
            u (int): vertex id

        Returns:
            bool: whether vertex `u` is a boundary vertex or not.
        """
        if self._is_vertex_on_border is None:
            self._compute_interior_boundary_vertices()
        return self._is_vertex_on_border[u]

    @property
    def interior_edges(self):
        if self._interior_edges is None:
            self._compute_interior_boundary_edges()
        return self._interior_edges

    @property
    def boundary_edges(self):
        if self._boundary_edges is None:
            self._compute_interior_boundary_edges()
        return self._boundary_edges

    @property
    def boundary_vertices(self):
        if self._boundary_vertices is None:
            self._compute_interior_boundary_vertices()
        return self._boundary_vertices
    
    @property
    def interior_vertices(self):
        if self._interior_vertices is None:
            self._compute_interior_boundary_vertices()
        return self._interior_vertices

    ###### Connectivity ######

    class _Connectivity(PolyLine._Connectivity):
        """Connectivity for surface meshes. Is accessed via the `.connectivity` attribute of the `SurfaceMesh` class.

        Warning:
            If your mesh is not manifold, there is no guarantee that the connectivity arrays will be correct.
        """

        def __init__(self, master):
            super().__init__(master)

            self._half_edges : dict = None # (v1,v2) -> (corner, previous, next, opposite, face, i1, i2)
            self._Cn2he      : dict = None # corner -> (v1,v2)
            self._adjVF2Cn   : dict = None # vertex,face -> corner
            self._adjV2Cn    : dict = None # vertex -> corners
            self._adjF2Cn    : dict = None # face -> corners
            self._face_id    : dict = None # (list of vertices) -> face
        
        def clear(self):
            """
            Resets connectivity. 
            The next query in the code will regenerate internal arrays.
            """
            super().clear()
            self._half_edges : dict = None # (v1,v2) -> (corner, previous, next, opposite, face, i1, i2)
            self._Cn2he      : dict = None # corner -> (v1,v2)
            self._adjVF2Cn   : dict = None # vertex,face -> corner
            self._adjV2Cn    : dict = None # vertex -> corners
            self._adjF2Cn    : dict = None # face -> first corner
            self._face_id    : dict = None # (list of vertices) -> face

        def _compute_face_ids(self):
            self._face_id = dict()
            for iF, F in enumerate(self.mesh.faces):
                self._face_id[utils.keyify(F)] = iF

        def face_id(self, *args) -> int:
            """ The identifier of a face.
            
            Args:
                int*: integers representing indices of vertices of the face (not necessarily in the correct order)

            Returns:
                int: A face index or None if the given tuple is invalid
            """
            if self._face_id is None:
                self._compute_face_ids()
            key = utils.keyify(*args)
            return self._face_id.get(key, None)

        def _compute_connectivity(self):
            super()._compute_connectivity()

            ### Compute corner connectivity dictionnaries
            self._adjV2Cn = dict([(i,set()) for i in self.mesh.id_vertices])
            self._adjVF2Cn = dict()
            self._adjF2Cn = dict()
            for iC in self.mesh.id_corners:
                v,f = self.mesh.face_corners.element(iC), self.mesh.face_corners.adj(iC)
                self._adjV2Cn[v].add(iC)
                self._adjVF2Cn[(v,f)] = iC
                if f not in self._adjF2Cn:
                    self._adjF2Cn[f] = iC
            for v in self.mesh.id_vertices:
                # recast sets as list for indexing and sorting
                self._adjV2Cn[v] = list(self._adjV2Cn[v])
                
            ### Compute half edges
            self._half_edges = dict()
            self._Cn2he = dict()
            for iF,F in enumerate(self.mesh.faces):
                n = len(F)
                for iV in range(n):
                    P, Pprev, Pnext = F[iV], F[(iV-1)%n], F[(iV+1)%n]
                    iC, iCprev, iCnext = (self._adjVF2Cn[(p,iF)] for p in (P,Pprev,Pnext)) 
                    self._half_edges[ (P, Pnext) ] = [iC, iCprev, iCnext, None, iF, iV, (iV+1)%n]
                    self._Cn2he[iC] = (P,Pnext)
            for (A,B) in self._half_edges.keys():
                iC1 = self._half_edges.get((A,B), [None])[0] # index of corner if it exists, else None 
                iC2 = self._half_edges.get((B,A), [None])[0]
                if iC1 is not None and iC2 is not None:
                    # iC1 and iC2 are opposite corners
                    self._half_edges[(A,B)][3] = iC2
                    self._half_edges[(B,A)][3] = iC1

            ### Sorting vertex neighborhoods
            if config.sort_neighborhoods and isinstance(self.mesh, SurfaceMesh):
                # Neighborhood should be sorted to have a clockwise order
                # Ignore sorting for volume meshes
                self._sort_vertex_neighborhoods()
            
        def _sort_vertex_neighborhoods(self):
            for A in self.mesh.id_vertices:
                corners_A = self._adjV2Cn[A]
                if len(corners_A)==0: continue
                sort_index = dict([(c,0) for c in corners_A])
                ind = 0
                is_boundary = False
                Cn = corners_A[0]
                for _ in range(len(sort_index)):
                    sort_index[Cn] = ind 
                    ind -= 1 
                    Cn = self.opposite_corner(self.previous_corner(Cn))
                    if Cn is None: 
                        is_boundary = True
                        break
                if is_boundary:
                    # also go counter clockwise 
                    ind = 0
                    Cn = corners_A[0]
                    for _ in range(len(sort_index)):
                        sort_index[Cn] = ind 
                        ind +=1
                        Cn = self.opposite_corner(Cn)
                        if Cn is None : break
                        Cn = self.next_corner(Cn)
                self._adjV2Cn[A].sort(key = lambda c : sort_index[c])
                sort_indexV = dict()
                for v in self._adjV2V[A]:
                    # all vertices have an associated corner except the last one (opposite half edge does not exist on boundary)
                    sort_indexV[v] = sort_index.get(self.half_edge_to_corner(A,v), -float("inf"))
                self._adjV2V[A].sort(key = lambda v : sort_indexV[v])

        ##### Vertices to elements

        def vertex_to_faces(self, V: int) -> list:
            """
            Neighborhood of vertex `V` in terms of faces.

            Args:
                V (int): vertex id

            Returns:
                list: list of faces `F` such that `V` is a vertex of `F`.
            """
            return [self.corner_to_face(iC) for iC in self.vertex_to_corners(V)]

        def vertex_to_corners(self, V: int) -> list:
            """
            List of face corners that correspond to vertex `V`

            Args:
                V (int): vertex id

            Returns:
                list: the list of corners `C` such that `mesh.corners[C]==V`
            """
            if self._adjV2Cn is None :
                self._compute_connectivity()
            return self._adjV2Cn.get(V, None)
        
        def vertex_to_corner_in_face(self, V: int, F: int) -> int:
            """
            The corner `C` corresponding to vertex `V` in face `F`.

            Args:
                V (int): vertex id
                F (int): face id

            Returns:
                int: corner id, or `None` if `V` is not a vertex of `F`.
            """
            if self._adjVF2Cn is None:
                self._compute_connectivity()
            return self._adjVF2Cn.get((V,F), None)
        
        #### Corner to element
        
        def previous_corner(self, C: int) -> int:
            """Previous corner of `C` around its associated face

            Args:
                C (int): corner index

            Returns:
                int: index of the previous corner
            """
            if self._half_edges is None:
                self._compute_connectivity()
            key = self._Cn2he.get(C,None)
            if key is None: return None
            return self._half_edges[key][1]

        def next_corner(self, C:int) -> int:
            """Next corner of `C` around its associated face

            Args:
                C (int): corner index

            Returns:
                int: index of the next corner
            """
            if self._half_edges is None:
                self._compute_connectivity()
            key = self._Cn2he.get(C,None)
            if key is None: return None
            return self._half_edges[key][2]

        def opposite_corner(self, C: int) -> int:
            """Opposite corner of `C` in terms of half edges. 
            If `C.vertex = A` and `C.next.vertex = B`, then returns the corner D such that `D.vertex = B` and `D.vertex.next = A`

            Args:
                C (int): corner index

            Returns:
                int: index of the opposite corner
            """
            if self._half_edges is None:
                self._compute_connectivity()
            key = self._Cn2he.get(C,None)
            if key is None: return None
            return self._half_edges[key][3]

        def corner_to_half_edge(self, C: int) -> int:
            if self._Cn2he is None:
                self._compute_connectivity()
            return self._Cn2he.get(C,None)

        def corner_to_face(self, C: int) -> int:
            """
            The face inside which corner `C` belongs.

            Args:
                C (int): corner id

            Returns:
                int: face id
            """
            return self.mesh.face_corners.adj(C)
        
        #### Edge to element
        
        def half_edge_to_corner(self, u: int, v: int) -> int:
            return self._half_edges.get((u,v), [None])[0]

        def direct_face(self, u: int, v: int, return_inds: bool = False):
            """Pair (u,v) of vertex -> triangle to the left of edge (u,v) if edge (u,v) exists, None otherwise
            Also returns local indexes of u and v in the triangle (and None if (u,v) does not exists)
            
            Calling this function with edge (v,u) yield the triangle of the other side of the edge

            Args:
                u (int): first vertex index
                v (int): second vertex index
                return_inds (bool, optional): Whether to return local indices of u and v in the face. Defaults to False.

            Returns:
                Index of a face or None. If return_inds is True, tuple made of index of face and local indices of u and v in face or (None,None,None). 
            """
            if self._half_edges is None:
                self._compute_connectivity()
            if (u,v) in self._half_edges:
                if return_inds: return self._half_edges[(u,v)][4:]
                return self._half_edges[(u,v)][4]
            else:
                if return_inds: return None,None,None
                return None

        def edge_to_faces(self, u: int, v: int):
            """For an edge `(u,v)`, returns the indices of faces that are direct and indirect from `(u,v)`(see `direct_face` method). If one of the faces does not exist, returns a `None` instead of its index.

            Args:
                u (int): index of first vertex
                v (int): index of second vertex
            """
            return self.direct_face(u,v), self.direct_face(v,u)

        def opposite_face(self, u : int, v : int, F : int, return_inds: bool = False):
            """Returns the index of the face adjacent to edge `(u,v)` that is at the opposite of face `F`. Returns `None` if such a face does not exist, or if `F` is not adjacent to edge `(u,v)`

            Args:
                u (int): index of edge's first vertex
                v (int): index of edge's second vertex
                F (int): index of face
                return_inds (bool, optional): If True, also returns local indices of `u` and `v` in opposite face. If input is invalid, will return `None,None,None` instead of `None`. Defaults to False.

            Returns:
                int: index of a face
            """
            if return_inds:
                F1, u1, v1 = self.direct_face(u,v, True)
                F2, v2, u2 = self.direct_face(v,u, True)
                if F1==F: return F2,u2,v2
                if F2==F: return F1,u1,v1
                return None,None,None
            else:
                F1, F2 = self.direct_face(u,v), self.direct_face(v,u)
                if F==F1: return F2
                if F==F2: return F1
                return None

        def common_edge(self, iF1: int, iF2: int):
            """Returns the two vertices (u,v) of the edge that separates faces iF1 and iF2 if it exists, and (None,None) otherwise.

            Args:
                iF1 (int): first face index
                iF2 (int): second face index

            Returns:
                (int,int): (u,v) pair of vertex indices, or (None,None)
            """
            F1 = self.mesh.faces[iF1]
            n = len(F1)
            for i in range(n):
                A,B = F1[i], F1[(i+1)%n]
                if self.opposite_face(A,B,iF1)==iF2:
                    return utils.keyify(A,B)
            return None,None

        ##### Face to element

        def face_to_vertices(self, F:int) -> list:
            """
            Neighborhood of face `F` in terms of vertices.

            Note:
                Equivalent to `mesh.faces[F]`

            Args:
                F (int): face id

            Returns:
                list: list of vertices `V` such that `V` is a vertex of `F`.
            """
            return list(self.mesh.faces[F])

        def in_face_index(self, F, V):
            """Index of vertex V in face F. None if V is not in face F

            Parameters:
                F (int): face index
                V (int): vertex index

            Returns:
                int
            """
            for (i,v) in enumerate(self.mesh.faces[F]):
                if v==V: return i
            return None
        
        def face_to_edges(self, F:int) -> list:
            """
            List of edges that bound face `F`.

            Args:
                F (int): face id

            Returns:
                list: list of edges `E` such that `E` is a boundary edge of face `F`
            """
            lF = self.mesh.faces[F]
            n = len(lF)
            return [self.edge_id(lF[i],lF[(i+1)%n]) for i in range(n)]

        def face_to_first_corner(self, F: int) -> int:
            """
            One corner `C` of the face `F` (the first in order of appearance in the `face_corners` container)

            Args:
                F (int): face id

            Returns:
                int: corner id
            """
            if self._adjF2Cn is None:
                self._compute_connectivity()
            return self._adjF2Cn[F]

        def face_to_corners(self, F: int) -> list:
            """
            list of corners of face `F`

            Args:
                F (int): face id

            Returns:
                list: list of corners of face `F`
            """
            if self._adjF2Cn is None:
                self._compute_connectivity()
            return [self._adjF2Cn[F] + _i for _i in range(len(self.mesh.faces[F]))]

        def face_to_faces(self, F:int) -> list:
            """
            List of faces that are adjacent to face `F`

            Args:
                F (int): face id

            Returns:
                list: list of faces `G` that are adjacent to `F`
            """
            if self._adjF2Cn is None:
                self._compute_connectivity()
            opposites = [self.opposite_corner(C) for C in self.face_to_corners(F)]
            return [self.corner_to_face(Op) for Op in opposites if Op is not None]


from .base import FrameField

from ...mesh.datatypes import *
from ...mesh.mesh_data import RawMeshData
from ...mesh.mesh_attributes import Attribute

from ... import attributes
from ..features import FeatureEdgeDetector

from ... import geometry as geom
from ...geometry import Vec, SphericalHarmonics
from ...geometry.rotations import match_rotation
from ...geometry import transform

from ...operators.laplacian_op import *
from ...procedural import axis_aligned_cube

import numpy as np
import scipy
import scipy.sparse as sp
from tqdm import tqdm
from scipy.spatial.transform import Rotation
from math import atan2
from osqp import OSQP
from ...utils.osqp_lin_solve import get_osqp_lin_solver

class FrameField3DVertices(FrameField): 

    @allowed_mesh_types(VolumeMesh)
    def __init__(self, 
        supporting_mesh : VolumeMesh, 
        feature_edges : bool = True, 
        verbose=True,
        **kwargs):
        super().__init__("vertices", "FrameField3D", verbose)
        self.mesh : VolumeMesh = supporting_mesh
        supporting_mesh.enable_boundary_connectivity()

        self.boundary_mesh : SurfaceMesh = None
        self.vertex_normals : Attribute = None
        self.n_smooth = kwargs.get("n_smooth", 3)
        self.smooth_attach_weight = kwargs.get("smooth_attach_weight", None)

        self.frames : list = None # raw representation of frames as scipy.Rotation objects
        self.var = np.concatenate([SphericalHarmonics.EYE]*len(self.mesh.vertices)) # representation vectors in spherical harmonics basis

        self._omegas : dict = None # rotation on edges
        self._singuls_tri : Attribute = None # value of defect on every triangle

        self.features : bool = feature_edges # whether feature edges are enabled or not
        self.feat : FeatureEdgeDetector = kwargs.get("custom_boundary_features", None)

    def initialize(self):
        self.log(" | Compute boundary manifold")
        self.boundary_mesh = self.mesh.boundary_connectivity.mesh
        if self.features and self.feat is None :
            self.feat = FeatureEdgeDetector(flag_corners=False, verbose=self.verbose)(self.boundary_mesh)
        self.vertex_normals = attributes.vertex_normals(self.boundary_mesh, persistent=False, interpolation="angle")
        
        self.log(" | Compute spherical harmonics coordinates on the boundary")
        face_bases_sh = Attribute(float, 9)
        for iF,F in enumerate(self.boundary_mesh.faces):
            pA,pB,pC = (self.boundary_mesh.vertices[v] for v in F)
            _,_,NF = geom.face_basis(pA,pB,pC)
            axis = geom.cross(Vec(0.,0.,1.), NF)
            if abs(NF.z)<0.99:
                # if abs(axis.norm()) > 1:
                #     axis = .99 * Vec.normalized(axis)
                axis = Vec.normalized(axis) * atan2(axis.norm(), NF.z)
                face_bases_sh[iF] = SphericalHarmonics.from_vec3(axis)
            else:
                face_bases_sh[iF] =  Vec(0., 0., 0., 0., 1., 0., 0., 0., 0.)
        vertex_bases_sh = Attribute(float, 9)
        vertex_bases_sh = attributes.interpolate_faces_to_vertices(self.boundary_mesh, face_bases_sh, vertex_bases_sh, weight="angle")

        self.frames = [Rotation.identity() for _ in self.mesh.id_vertices]
        for vb in self.boundary_mesh.id_vertices:
            v = self.mesh.boundary_connectivity.b2m_vertex[vb]
            f,a = SphericalHarmonics.project_to_frame(vertex_bases_sh[vb])
            self.var[9*v:9*(v+1)] = a
            self.frames[v] = f
        self.initialized = True

    def _compute_frame_from_sh(self):
        iterator = self.mesh.id_vertices
        if self.verbose: iterator = tqdm(iterator, desc="sh_to_frames")
        for v in iterator:
            shv = Vec(self.var[9*v:9*(v+1)])
            f,a = SphericalHarmonics.project_to_frame(shv)
            self.var[9*v:9*(v+1)] = a
            self.frames[v] = f

    def _compute_sh_from_frames(self):
        for i, frame in enumerate(self.frames):
            self.var[9*i:9*(i+1)] = SphericalHarmonics.from_frame(frame)

    def _graph_laplacian(self):
        """Tensor product between graph laplacian matrix and Id(9x9)"""
        lap = graph_laplacian(self.mesh).tolil()
        return scipy.sparse.kron(lap, scipy.sparse.eye(9), format="csc")

    def _volume_laplacian(self):
        lap = volume_laplacian(self.mesh).tolil()
        return scipy.sparse.kron(lap, scipy.sparse.eye(9), format="csc")

    def _volume_weight_matrix(self):
        A = volume_weight_matrix(self.mesh).tolil()
        return scipy.sparse.kron(A, scipy.sparse.eye(9), format="csc")

    def compute_constraints(self):
        self.log(" | Compute constraints matrix")
        nvar = len(self.var)
        ncstr_normals = len(self.mesh.boundary_vertices) - (0 if self.feat is None else len(self.feat.feature_vertices))
        ncstr_features = 0 if self.feat is None else 9*len(self.feat.feature_vertices)
        ncstr = ncstr_normals + ncstr_features
        cstrMat = scipy.sparse.lil_matrix((ncstr, nvar))
        cstrRHS = np.zeros(ncstr)
        row = 0
        for iv in self.boundary_mesh.id_vertices:
            if self.feat is not None and iv in self.feat.feature_vertices: continue
            axis = rotations.axis_rot_from_z(self.vertex_normals[iv])
            nrml_frame = SphericalHarmonics.from_vec3(axis)
            v = self.mesh.boundary_connectivity.b2m_vertex[iv]
            for _c in range(9):
                cstrMat[row,9*v+_c] = nrml_frame[_c]
            cstrRHS[row] = nrml_frame.dot(self.var[9*v:9*(v+1)])
            row += 1

        if self.feat is not None:
            for iv in self.feat.feature_vertices:
                v = self.mesh.boundary_connectivity.b2m_vertex[iv]
                # frame at iv is locked
                for _c in range(9):
                    cstrMat[row+_c, 9*v+_c] = 1
                    cstrRHS[row+_c] = self.var[9*v+_c]
                row += 9
        cstrMat = cstrMat.tocsc()
        return cstrMat, cstrRHS

    def normalize(self):
        """Calls projection for every vertex and also updates representations variables. 
        The projection garantees that boundary frames stay aligned with their normal.
        """
        if self.feat is None:
            iterator = self.mesh.id_vertices
        else:
            iterator = (v for v in self.mesh.id_vertices if not self.mesh.is_vertex_on_border(v) or self.mesh.boundary_connectivity.m2b_vertex[v] not in self.feat.feature_vertices)
        if self.verbose: iterator = tqdm(iterator, desc="normalize")
        for i in iterator:
            nrml=self.vertex_normals[i] if self.mesh.is_vertex_on_border(i) else None
            frame, a = SphericalHarmonics.project_to_frame(self.var[9*i:9*(i+1)], stop_threshold=1e-5, nrml_cstr=nrml)
            self.var[9*i:9*(i+1)] = a
            self.frames[i] = frame

    def optimize(self):
        self._check_init()
        # Create Laplacian
        lap = self._volume_laplacian()
        # lap = self._graph_laplacian()
        A = self._volume_weight_matrix()

        # Create Constraints Matrix
        nvar = len(self.var)
        cstrMat, cstrRHS = self.compute_constraints()
       
        self.log(" | Initial solve of linear system")
        instance = OSQP()
        instance.setup(lap, None, A=cstrMat, l=cstrRHS, u=cstrRHS, verbose=self.verbose, linsys_solver=get_osqp_lin_solver())
        res = instance.solve()
        self.var = res.x

        if self.smooth_attach_weight is not None:
            alpha = self.smooth_attach_weight
        else:
            try:
                alpha = abs(sp.linalg.eigsh(lap, k=1, which="SM", M=A, tol=1e-4, maxiter=1000, return_eigenvectors=False)[0])
            except Exception as e:
                self.log("Estimation of alpha failed: {}".format(e))
                alpha = abs(sp.linalg.eigsh(lap-0.01*sp.identity(lap.shape[0]), k=1, M=A, which="SM", tol=1e-4, maxiter=1000, return_eigenvectors=False)[0])            
        self.log("Attach weight: {}".format(alpha))

        if self.n_smooth>0:
            self.log("Solve linear system {} times with diffusion".format(self.n_smooth))
            Id = scipy.sparse.eye(nvar, format="csc")
            mat = lap + alpha * Id
            instance = OSQP()
            instance.setup(mat,self.var, A=cstrMat, l=cstrRHS, u=cstrRHS)
            for _ in range(self.n_smooth):
                self.normalize()
                instance.update(q = -alpha * self.var)
                res = instance.solve()
                self.var = res.x
        self.normalize()
        self.smoothed = True

    def flag_singularities(self):
        self._check_init()
        ZERO_THRESHOLD = 1e-2
        self.log("Flag Singularities")
        if self.mesh.faces.has_attribute("singuls"):
            self._singuls_tri = self.mesh.faces.attribute("singuls")
        else:
            self._singuls_tri = self.mesh.faces.create_attribute("singuls", float)
        self._omegas = dict()
        # first compute every matching for every edge
        for e,(A,B) in enumerate(self.mesh.edges):
            rA,rB = self.frames[A], self.frames[B]
            self._omegas[e] = match_rotation(rA,rB)

        for T in self.mesh.id_faces:
            A,B,C = self.mesh.faces[T]
            eAB,eBC,eCA = self.mesh.connectivity.edge_id(A,B), self.mesh.connectivity.edge_id(B,C), self.mesh.connectivity.edge_id(C,A)

            rAB = self.omegas[eAB] if A<B else self.omegas[eAB].inv()
            rBC = self.omegas[eBC] if B<C else self.omegas[eBC].inv()
            rCA = self.omegas[eCA] if C<A else self.omegas[eCA].inv()
            rot =  rCA * rBC * rAB
            dfct = rot.magnitude()
            if dfct> ZERO_THRESHOLD:
                self._singuls_tri[T] = dfct #round(dfct*2/pi)

    @property
    def omegas(self):
        if self._omegas is None:
            self.flag_singularities()
        return self._omegas

    @property
    def singuls(self):
        if self._singuls_tri is None:
            self.flag_singularities()
        return self._singuls_tri

    @property
    def singularity_graph(self):
        singu_graph = PolyLine()
        singuls_on_graph = singu_graph.edges.create_attribute("singus", float)
        bary_tet = attributes.cell_barycenter(self.mesh, persistent=False)
        bary_tri = attributes.face_barycenter(self.mesh, persistent=False)

        n = 0
        for Tri in self.mesh.id_faces:
            if abs(self.singuls[Tri])>1e-2:
                # triangle is singular
                if self.mesh.is_face_on_border(Tri):
                    Tet = self.mesh.connectivity.face_to_cells(Tri)[0]
                    singu_graph.vertices += [bary_tri[Tri], bary_tet[Tet]]
                else:
                    T1, T2 = self.mesh.connectivity.face_to_cells(Tri)
                    singu_graph.vertices += [bary_tet[T1], bary_tet[T2]]
                singu_graph.edges.append((2*n, 2*n+1))
                singuls_on_graph[n] = self.singuls[Tri]
                n += 1
        return singu_graph

    def export_as_mesh(self) -> SurfaceMesh:
        """
        Exports the frame field as a mesh for visualization.

        Returns:
            SurfaceMesh: the frame field as a mesh object, where small cubes represent frames
        """
        self._check_init()
        L = attributes.mean_edge_length(self.mesh,50)
        FFMesh = RawMeshData()
        col = FFMesh.faces.create_attribute("color", float, 3)
        RED,GREEN,BLUE = Vec(1.,0.,0), Vec(0.,1.,0.), Vec(0.,0.1,1.)
        
        for i,pV in enumerate(self.mesh.vertices):
            r : Rotation = self.frames[i]
            cube_v = axis_aligned_cube()
            cube_v = transform.scale(cube_v, 0.5*L, orig=Vec(0.,0.,0.))
            cube_v = transform.rotate(cube_v, r, orig=Vec(0.,0.,0.))
            cube_v = transform.translate(cube_v, pV)
            FFMesh.vertices += cube_v.vertices
            FFMesh.faces += [tuple((8*i+x for x in F)) for F in cube_v.faces]

            col[12*i] = RED
            col[12*i+1] = RED
            col[12*i+10] = RED
            col[12*i+11] = RED

            col[12*i+2] = GREEN
            col[12*i+3] = GREEN
            col[12*i+6] = GREEN
            col[12*i+7] = GREEN
            
            col[12*i+4] = BLUE
            col[12*i+5] = BLUE
            col[12*i+8] = BLUE
            col[12*i+9] = BLUE
        return SurfaceMesh(FFMesh)



from .base import BaseParametrization

from ...mesh.datatypes import SurfaceMesh,PolyLine
from ...mesh.mesh_attributes import Attribute
from ...mesh.mesh import copy

from ..cutting import SingularityCutter
from ..connection import SurfaceConnectionFaces
from ..features import FeatureEdgeDetector
from ..framefield import FrameField
from ..trees import FaceSpanningTree
from ...import attributes

from ... import geometry as geom
from ...geometry import Vec
from ... import utils

import numpy as np
import cmath
import scipy.sparse as sp
from osqp import OSQP

class FrameFieldIntegration(BaseParametrization):
    """
    Integration of a smooth vector field into a seamless parametrization

    References:
        - [1] _Mixed-integer quadrangulation_, Bommes D, Zimmer H. and Kobbelt L., ACM Transaction on Graphics, 2009
        
        - [2] _QuadCover - Surface Parameterization using Branched Coverings_, Kälberer F., Nieser M. and Polthier K., Computer Graphics Forum, 2007
    """

    def __init__(self, framefield: FrameField, verbose:bool=False, **kwargs):
        """
        Args:
            framefield (FrameField): the frame field object to integrate.
            singularities (Attribute): float Attribute on vertices. Gives the target angle defects of vertices
            use_cotan (bool, optional): if True, uses cotangents as weights in the laplacian matrix. Defaults to True. 
            verbose (bool, optional): verbose mode. Defaults to False.

        Keyword Args:    
            debug (bool, optional) : debug mode. Generates additionnal output. Defaults to False.

        Note:
            It is not necessary to provide the input mesh, as it is accessed via the `framefield` argument

        Example:
            [https://github.com/GCoiffier/mouette/blob/main/examples/parametrization/ff_integrate.py](https://github.com/GCoiffier/mouette/blob/main/examples/parametrization/ff_integrate.py)
        
        Raises:
            AssertionError : the frame field should be defined on faces and be of order 4
        """
        super().__init__("Frame Field Integration", framefield.mesh, verbose, **kwargs)
        self._use_cotan = kwargs.get("use_cotan", True)
        self._debug = kwargs.get("debug", False)
        self.save_on_corners = True

        self.frame_field = framefield
        assert self.frame_field.element == "faces"
        assert self.frame_field.order == 4

        self._cutter : SingularityCutter
        self._matching : np.ndarray = None # matching attribute on faces
        self._jumps : Attribute = None # jump attribute on edges
        self._singus : Attribute = None # position and indices of cones (on vertices)
        self._tree : FaceSpanningTree = None # traversal tree

    @property
    def _conn(self) -> SurfaceConnectionFaces:
        # Shortcut getter
        return self.frame_field.conn

    @property
    def _feat(self) -> FeatureEdgeDetector:
        # Shortcut getter
        return self.frame_field.feat

    def _get_ff_dir(self, T:int) -> complex:
        """
        First frame field direction after the matching was computed

        Args:
            T (int): face index

        Returns:
            complex : the matched frame branch  
        """
        if self._matching is None: return None
        return utils.maths.roots(self.frame_field[T],4)[self._matching[T]]

    def _compute_matching(self):
        self._tree = FaceSpanningTree(self.mesh, forbidden_edges=self._cutter.cut_edges)()
        self._matching = np.zeros(len(self.mesh.faces),dtype=np.int32)
        for face, parent in self._tree.traverse():
            if parent is None: continue # brushing of root is 0 by convention
            rp = self._get_ff_dir(parent)
            angles = [utils.maths.angle_diff( cmath.phase(rp), cmath.phase(rf) - self._conn.transport(face,parent)) for rf in utils.maths.roots(self.frame_field[face], 4)]
            abs_angles = [abs(_a) for _a in angles]
            self._matching[face] = np.argmin(abs_angles)

    def _compute_jumps(self):
        self.jumps = self.mesh.edges.create_attribute("jump", int)
        for e in self._cutter.cut_edges:
            T1,T2 = self.mesh.connectivity.edge_to_faces(*self.mesh.edges[e])
            if T1 is None or T2 is None : continue # boundary edge -> no jump
            ff1, ff2 = self._get_ff_dir(T1), self._get_ff_dir(T2)
            jump = utils.maths.angle_diff( cmath.phase(ff1), cmath.phase(ff2) - self._conn.transport(T2,T1))
            self.jumps[e] = round(2*jump/np.pi)%4 # integer multiple of pi/2

    def _integrate(self):
        nvar = 4*len(self.mesh.faces) # one 2x2 Jacobian per face
        
        ## Compute objective
        q = np.zeros(nvar, dtype=float)
        for T in self.mesh.id_faces:
            # We want the jacobian ( a  b ) to match the rotation matrix of the frame ( Re(zT)  Im(zT) )
            #                      ( c  d )                                           ( -Im(zT) Re(zT) )
            zT = self._get_ff_dir(T)
            q[4*T:4*T+4] = [zT.real, zT.imag, -zT.imag, zT.real]
        
        ## Compute constraints
        rows,cols,vals = [], [], []
        irow = 0
        
        ## Build jump constraint matrix
        for e in self.mesh.id_edges:
            je = self.jumps[e]
            A,B = self.mesh.edges[e]
            T1,T2 = self.mesh.connectivity.edge_to_faces(A,B)
            if T1 is None or T2 is None : continue # boundary edge
            E = self.mesh.vertices[B] - self.mesh.vertices[A]
            E1 = self._conn.project(E,T1)
            E2 = self._conn.project(E,T2)
            
            # J1 E1 = R^{je} J2 E2
            rows += [irow, irow,   irow+1, irow+1]
            cols += [4*T1, 4*T1+1, 4*T1+2, 4*T1+3]
            vals += [E1.x, E1.y, E1.x, E1.y]

            rows += [irow, irow,   irow+1, irow+1]
            if je==0:
                cols += [4*T2, 4*T2+1, 4*T2+2, 4*T2+3]
                vals += [-E2.x, -E2.y, -E2.x, -E2.y]
            elif je==1:
                cols += [4*T2+2, 4*T2+3, 4*T2, 4*T2+1]
                vals += [-E2.x, -E2.y, E2.x, E2.y]
            elif je==2:
                cols += [4*T2, 4*T2+1, 4*T2+2, 4*T2+3]
                vals += [E2.x, E2.y, E2.x, E2.y]
            elif je==3:
                cols += [4*T2+2, 4*T2+3, 4*T2, 4*T2+1]
                vals += [E2.x, E2.y, -E2.x, -E2.y]
            irow += 2
        
        ## Feature edge alignment
        for e in self._feat.feature_edges:
            A,B = self.mesh.edges[e]
            E = self.mesh.vertices[B] - self.mesh.vertices[A] 
            for T in self.mesh.connectivity.edge_to_faces(A,B):
                if T is None: continue
                ET = Vec.normalized(self._conn.project(E,T))
                zT = self._get_ff_dir(T)
                zT = Vec(zT.real, zT.imag)
                # determine the orientation of the feature (horizontal or vertical) in parameter space depending on the matching
                dtp = round(abs(ET.dot(zT)))
                rows += [irow, irow]
                irow += 1
                if dtp==0: # zero => horizontal
                    cols += [4*T,4*T+1]
                    vals += [ET.x, ET.y]
                elif dtp==1:
                    cols += [4*T+2,4*T+3]
                    vals += [ET.x, ET.y]

        CstMat = sp.csc_matrix((vals, (rows,cols)), shape=(irow,nvar))
        CstRHS = np.zeros(irow)

        ## Solve system
        instance = OSQP()
        instance.setup(sp.eye(nvar,format="csc"), -q, CstMat, CstRHS, CstRHS, verbose=self.verbose)
        result = instance.solve()
        return result.x

    def _export_jacobian_dets(self,Jac):
        det_attr = self.mesh.faces.create_attribute("det", float)
        for T in self.mesh.id_faces:
            J = Jac[4*T:4*T+4].reshape((2,2))
            detJ = np.linalg.det(J)
            det_attr[T] = detJ

    def _build_UVs(self,Jac):
        self.uvs = self.mesh.face_corners.create_attribute("uv_coords", float, 2, dense=True)

        def build_triangle(iT):
            pA,pB,pC = (self.mesh.vertices[_u] for _u in self.mesh.faces[iT])
            X,Y = self._conn.base(iT)
            qA = Vec.zeros(2)
            qB = Vec(X.dot(pB-pA), Y.dot(pB-pA))
            qC = Vec(X.dot(pC-pA), Y.dot(pC-pA))
            J = Jac[4*iT:4*iT+4].reshape((2,2))
            qB = Vec(J.dot(qB))
            qC = Vec(J.dot(qC))
            return qA,qB,qC
        
        def align_triangles(pA,pB, qA, qB, qC):
            target = Vec(pB - pA)

            # 1) translation to align point A
            translation = pA - qA
            qA += translation
            qB += translation
            qC += translation

            # 2) rotation around point A to align the point B
            q = Vec(qB - qA)
            rot = np.arctan2(target.y, target.x) - np.arctan2(q.y,q.x)
            rc, rs = np.cos(rot), np.sin(rot)

            q = Vec(qB - qA)
            qB.x, qB.y = qA.x + rc*q.x - rs*q.y , qA.y + rs*q.x + rc*q.y
            q = Vec(qC-qA)
            qC.x, qC.y = qA.x + rc*q.x - rs*q.y , qA.y + rs*q.x + rc*q.y

            return qA,qB,qC

        for T,parent in self._tree.traverse():
            if parent is None:
                # build the root
                pA,pB,pC = build_triangle(T)
                self.uvs[3*T+0] = pA
                self.uvs[3*T+1] = pB
                self.uvs[3*T+2] = pC
                continue

            A,B = self.mesh.connectivity.common_edge(T,parent) # common vertices
            iA,iB = (self.mesh.connectivity.in_face_index(parent,_u) for _u in (A,B))
            jA,jB = (self.mesh.connectivity.in_face_index(T,_u) for _u in (A,B))
            jC = 3-jA-jB
            
            pA,pB = self.uvs[3*parent+iA], self.uvs[3*parent+iB]
            q = build_triangle(T)
            qA,qB,qC = q[jA], q[jB], q[jC]
            qA,qB,qC = align_triangles(pA,pB,qA,qB,qC)

            for (j,q) in [(jA,qA), (jB,qB), (jC,qC)]:
                self.uvs[3*T+j] = q

    def run(self):
        """Computes the integration"""

        self.log("Smooth frame field if needed")
        self.frame_field.run()
        self.frame_field.flag_singularities() # computes singularities
        self._singus = self.mesh.vertices.get_attribute("singuls")

        self.log("Perform cuts between cones")
        ## Define a cut mesh along the shortest path towards the boundary
        self._cutter = SingularityCutter(self.mesh, self._singus, features=self._feat)
        self._cutter.run()

        self.log("Compute frame field matching")
        self._compute_matching()
        self._compute_jumps()
       
        self.log("Solve gradient least-square system")
        Jac = self._integrate()
        self._export_jacobian_dets(Jac)

        self.log("Build UVs")
        self._build_UVs(Jac)

        self.log("Rescale UVs and output result")
        ## Rescale UV coordinates to fit in [0,1]²
        xmin,xmax,ymin,ymax = float("inf"), -float("inf"), float("inf"), -float("inf")
        for c in self.mesh.id_corners:
            cx,cy = self.uvs[c]
            xmin = min(xmin, cx)
            xmax = max(xmax, cx)
            ymin = min(ymin, cy)
            ymax = max(ymax, cy)
        scale_x = xmax-xmin
        scale_y = ymax-ymin
        scale = min(scale_x, scale_y)

        ## Write final result in attribute
        for c in self.mesh.id_corners:
            self.uvs[c] = self.uvs[c]/scale

    def export_frame_field_as_mesh(self) -> PolyLine:
        """
        Exports the frame field as a PolyLine for visualization
        """
        if self._matching is None:
            return self.frame_field.export_as_mesh()

        FFmesh = PolyLine()
        L = attributes.mean_edge_length(self.mesh)/3
        match_attr = FFmesh.edges.create_attribute("matching", int, dense=True)
        for id_face, face in enumerate(self.mesh.faces):
            basis,Y = self._conn.base(id_face)
            normal = geom.cross(basis,Y)
            pA,pB,pC = (self.mesh.vertices[_v] for _v in face)
            angle = cmath.phase(self._get_ff_dir(id_face))
            bary = (pA+pB+pC)/3 # reference point for display
            r1,r2 = (geom.rotate_around_axis(basis, normal, angle + k*np.pi/2) for k in range(2))
            p1,p2= (bary + L*r for r in (r1,r2))
            FFmesh.vertices += [bary, p1, p2]
            FFmesh.edges += [(3*id_face, 3*id_face+k) for k in range(1,3)]
            match_attr[2*id_face]   = 0
            match_attr[2*id_face+1] = 1
        return FFmesh  

    @property
    def cut_mesh(self) -> SurfaceMesh:
        """Disk-topology mesh where cuts have been performed on seams"""
        return self._cutter.output_mesh

    @property
    def cut_graph(self) -> PolyLine:
        """Seam edges as a PolyLine"""
        return self._cutter.cut_graph

    @property
    def flat_mesh(self) -> SurfaceMesh:
        """
        A flat representation of the mesh where uv-coordinates are copied to xy.

        Returns:
            SurfaceMesh: the flat mesh
        """
        if self.uvs is None : return None
        if self._flat_mesh is None:
            # build the flat mesh : vertex coordinates are uv of original mesh
            self._flat_mesh = copy(self.cut_mesh)
            for T in self.cut_mesh.id_faces:
                for i,v in enumerate(self.cut_mesh.faces[T]):
                    self._flat_mesh.vertices[v] = Vec(self.uvs[3*T+i][0], self.uvs[3*T+i][1], 0.)
        return self._flat_mesh


from .. import geometry as geom
from ..mesh.datatypes import *
from ..utils import PriorityQueue
from ..utils.argument_check import *
from ..mesh.mesh_attributes import _BaseAttribute

def build_path(mesh : Mesh, paths):
    path_mesh = PolyLine()
    k = 0
    for l in paths.values():
        if len(l)>0:
            path_mesh.vertices.append(mesh.vertices[l[0]])
        if len(l)>1:
            for i in range(1, len(l)):
                path_mesh.vertices.append(mesh.vertices[l[i]])
                path_mesh.edges.append((k+i-1,k+i))
    return path_mesh

def _check_weight_argument(weights):
    if isinstance(weights, str):
        check_argument("weights", weights, str, ["one", "length"])
    elif not (isinstance(weights,dict) or isinstance(weights, _BaseAttribute)):
        raise InvalidArgumentTypeError("weights", type(weights), str, dict, _BaseAttribute)

@forbidden_mesh_types(PointCloud)
def shortest_path(mesh : Mesh, start : int, targets : list, weights = "length", export_path_mesh = False) -> dict:
    """
    Computes the shortest path from 'start' vertex to all given targets.
    Uses Dijsktra's algorithm

    Parameters:
        mesh (Union[Polyline, SurfaceMesh, VolumeMesh]): The mesh
        start (int): Vertex index of starting point on the mesh
        targets (int | list | set): 
            Vertex index of the end point on the mesh or
            list of vertex indexes for ending points on the mesh
        weights (str | dict) : provided weights of each edge. Options are:
            - "one" : uniform weight = 1 for every edge
            - "length" : use the length of the edge
            - any dict : custom weights
            Defaults to "length".
        export_path_mesh (bool, optional): 
            If specified, will also return the path as a Polyline file. 
            Defaults to False.
    
    Returns:
        dict: target_id -> list of vertices to visit on shortest path from 'start' to 'target_id'
        If export_path_mesh is set to True, also returns a Polyline
    """

    if isinstance(targets, int):
        targets = {targets}
    else:
        targets = set(targets)

    _check_weight_argument(weights)

    # Initialize data
    if weights=="one":
        edge_length = lambda _ : 1.
    elif weights== "length":
        edge_length = lambda u,v : geom.distance(mesh.vertices[u], mesh.vertices[v])
    else:
        edge_length = lambda u,v : weights[mesh.connectivity.edge_id(u,v)]

    queue = PriorityQueue()
    visited = dict([(_i,False) for _i in mesh.id_vertices])
    path = dict([(_i,None) for _i in mesh.id_vertices])
    distance = dict([(_i,float("inf")) for _i in mesh.id_vertices])
    distance[start] = 0.

    # Run Dijkstra's algorithm
    queue.push(start, 0.)
    while not queue.empty():
        v = queue.get().x
        if visited[v] : continue
        visited[v] = True
        for nv in mesh.connectivity.vertex_to_vertices(v):
            d = distance[v] + edge_length(v,nv)
            if distance[nv] > d:
                distance[nv] = d
                path[nv] = v
            if not visited[nv]:
                queue.push(nv, distance[nv])
    
    # Build paths
    paths_list = dict([(t, []) for t in targets])
    for t in targets:
        v = t
        while v != start:
            paths_list[t].append(v)
            v = path[v]
        paths_list[t].append(start)
        paths_list[t].reverse()
    
    if export_path_mesh:
        path_mesh = build_path(mesh, paths_list)
        return paths_list, path_mesh
    return paths_list

@forbidden_mesh_types(PointCloud)
def shortest_path_to_vertex_set(mesh : PolyLine, start : int, targets : list, weights = "length", export_path_mesh = False):
    """Computes the shortest path from 'start' vertex to the closest vertex in the target set
    The idea is to add fictionnal edges between all vertices of targets to a representent vertex, with weight 0, and call Dijsktra's algorithm to reach this vertex.

    Parameters:
        mesh (Union[Polyline, SurfaceMesh, VolumeMesh]): The mesh
        start (int): Vertex index of starting point on the mesh
        targets (list): list of vertices to reach
        weights (str | dict) : provided weights of each edge. Options are:
            - "one" : uniform weight = 1 for every edge
            - "length" : use the length of the edge
            - any dict : custom weights
            Defaults to "length".
        export_path_mesh (bool, optional): 
            If specified, will also return the path as a Polyline file. 
            Defaults to False.
    
    Raises:
        Exception: No target provided if the target list is empty

    Returns:
        int: the index of the closest vertex from the targets set
        list: the list of vertices on the closest path
        If export_path_mesh is set to True, also returns a Polyline
    """
    _check_weight_argument(weights)

    if not isinstance(targets, list):
        targets = [x for x in targets]

    if len(targets)==0 :
        raise Exception("No target provided")

    TARGET = -1 # an additionnal vertex linked to every vertex in the target set

    if len(targets)==1 : 
        # nothing special to do in this case, just call shortest_path between two vertices
        if export_path_mesh:
            parent, mesh = shortest_path(mesh, start, TARGET, weights, export_path_mesh)
            return TARGET, parent[TARGET], mesh
        else:
            parent = shortest_path(mesh, start, TARGET, weights, export_path_mesh)[TARGET]
            return TARGET, parent

    # Initialize data
    # build a dict u -> (v -> d) with u and v vertices and d the weight between them
    connectivity = dict([ (u, dict()) for u in mesh.id_vertices ])
    connectivity[TARGET] = dict()
    
    if weights=="one":
        for (u,v) in mesh.edges:
            connectivity[u][v] = 1
            connectivity[v][u] = 1

    elif weights== "length":
        for (u,v) in mesh.edges:
            d = geom.distance(mesh.vertices[u], mesh.vertices[v])
            connectivity[u][v] = d
            connectivity[v][u] = d
    else:
        for e,(u,v) in enumerate(mesh.edges):
            connectivity[u][v] = weights[e]
            connectivity[v][u] = weights[e]

    for s in targets:
        # TARGET is targets[0]
        connectivity[s][TARGET] = 0
        connectivity[TARGET][s] = 0

    queue = PriorityQueue()
    visited = dict([(_i,False) for _i in mesh.id_vertices])
    parent = dict([(_i,None) for _i in mesh.id_vertices])
    distance = dict([(_i,float("inf")) for _i in mesh.id_vertices])
    visited[TARGET], parent[TARGET], distance[TARGET] = False, None, float("inf")
    distance[start] = 0.

    # Run Dijkstra's algorithm
    queue.push(start, 0.)
    while not queue.empty():
        v = queue.get().x
        if visited[v] : continue
        visited[v] = True
        for nv in connectivity[v]:
            d = distance[v] + connectivity[v][nv]
            if distance[nv] > d:
                distance[nv] = d
                parent[nv] = v
            if not visited[nv]:
                queue.push(nv, distance[nv])
    
    # Build paths
    path = []
    v = TARGET
    while v != start:
        v = parent[v]
        path.append(v)
    path.reverse()
    ind = start if not path else path[-1]

    if export_path_mesh:
        path_mesh = build_path(mesh, {TARGET : path})
        return ind, path, path_mesh

    return ind, path

@allowed_mesh_types(SurfaceMesh)
def shortest_path_to_border(mesh : SurfaceMesh, start : int, weights = "length", export_path_mesh = False) -> list:
    """
    Computes the shortest path from 'start' vertex to the boundary of the mesh.
    Call to shortest_path_to_vertex_set with the set of boundary vertices

    Parameters:
        mesh (SurfaceMesh): The mesh
        start (int): Vertex index of starting point on the mesh
        weights (str | dict | Attribute) : provided weights of each edge. Options are:
            - "one" : uniform weight = 1 for every edge
            - "length" : use the length of the edge
            - any dict or Attribute on edges : custom weights
            weights are set to 0 for edges on the boundary. Defaults to "length".
        export_path_mesh (bool, optional): 
            If specified, will also return the path as a Polyline file. 
            Defaults to False.
    Raises:
        Exception: "Mesh has no border"
            raised if the border of the mesh does not exist

    Returns:
       list: The list of vertices on shortest path
       If export_path_mesh is set to True, also returns a Polyline
    """
    if len(mesh.boundary_vertices)==0:
        raise Exception("Mesh has no border")
    
    result = shortest_path_to_vertex_set(mesh, start, mesh.boundary_vertices, weights, export_path_mesh)
    if len(result)==2: # ignore first argument (id of vertex)
        return result[1]
    return result[1:]
    


from .vector import Vec
from ..utils import check_argument
import numpy as np
import math
from typing import Union

def sign0(x: float):
    """
    $$\\text{sign0}(x)=\\begin{cases} 
        1 \\text{ if } x\\geqslant0 \\\\
        -1 \\text{ if } x<0
    \\end{cases}$$

    Parameters:
        x (float): input number

    Returns:
        float: the sign of x
    """
    if x>=0: return 1
    return -1

def sign(x: float):
    """
    $$\\text{sign}(x)=\\begin{cases} 
        0 \\text{ if } x=0 \\\\
        1 \\text{ if } x>0 \\\\
        -1 \\text{ if } x<0
    \\end{cases}$$

    Parameters:
        x (float): input number

    Returns:
        float: x/|x| and 0 for x=0
    """
    if x>0: return 1
    elif x<0 : return -1
    return 0

def norm(x: np.ndarray, which="l2") -> float:
    """Vector norm. Three norms are implemented, the Euclidean l2 norm, the l1 norm or the l-infinite norm:

    l2: $ \\sqrt{ \\sum_i v[i]^2 } $

    l1: $ \\sum_i |v[i]| $

    linf: $ \\max_i |v[i]| $

    Args:
        x (np.ndarray): vector from which we compute the norm. Will be flattened if it has more than one dimension.
        which (str, optional): which norm to compute. Choices are "l2", "l1" and "linf". Defaults to "l2".

    Returns:
        float: the vector's norm
    """
    check_argument("which", which, str, ["l2", "l1", "linf"])
    if which=="l2":
        return np.sqrt(np.dot(x.flatten(), x.flatten()))
    elif which=="l1":
        return np.sum(np.abs(x))
    elif which=="linf":
        return np.max(np.abs(x))

def dot(A: np.ndarray, B: np.ndarray) -> float:
    """dot product $A^TB$ between two vectors

    Args:
        A (np.ndarray): first vector
        B (np.ndarray): second vector

    Returns:
        float: dot product
    """
    return np.dot(A,B)

def distance(A: Vec, B: Vec, which="l2") -> float:
    """Distance between A and B. Three distances are implemented, the Euclidean distance l2, the Manhattan l1 distance of the l-infinity distance:
     
    l2: $\\sqrt{(B-A)^T(B-A)} = \\sqrt{\\sum_i (A_i - B_i)^2}$

    l1: $\\sum_i |A_i - B_i|$

    linf: $\\max_i |A_i - B_i|$

    Args:
        A (Vec): first point
        B (Vec): second point
        which (str, optional): which norm to compute. Choices are "l2", "l1" and "linf". Defaults to "l2".
        
    Returns:
        float: distance
    """
    return norm(B-A, which)

def cross(A: Vec, B: Vec) -> Vec:
    """Cross product of vectors A and B

    Args:
        A (Vec): Size 3 vector
        B (Vec): Size 3 vector
        
    Returns:
        Vec: cross product AxB
    """
    return Vec(
        A[1]*B[2] - A[2] * B[1],
        B[0]*A[2] - B[2] * A[0],
        A[0]*B[1] - A[1] * B[0]
    )

def cotan(A: Vec, B: Vec, C: Vec) -> float:
    """cotangent of the angle $\\hat{ABC}$
    
    Args:
        A (Vec): point A
        B (Vec): point B
        C (Vec): point C

    Returns:
        float: the cotangent
    """
    A, B, C = Vec(A), Vec(B), Vec(C)
    BA = Vec.normalized(A-B)
    BC = Vec.normalized(C-B)
    cosine = np.dot(BA,BC)
    sine = norm(cross(BA,BC))
    return cosine/sine

def angle_3pts(A: Vec, B: Vec, C: Vec) -> float:
    """Angle ABC between three points

    Args:
        A (Vec): first point
        B (Vec): central point
        C (Vec): second point

    Returns:
        float: the angle
    """
    BA = Vec(A)-Vec(B)
    BC = Vec(C)-Vec(B)
    s = cross(BA,BC).norm()
    c = dot(BA,BC)
    return math.atan2(s, c)


def signed_angle_2vec3D(V1: Vec, V2: Vec, N: Vec) -> float:
    """Signed angle between two vectors with orientation given by normal N

    Args:
        V1 (Vec): First vector
        V2 (Vec): Second vector
        N (Vec): reference normal direction

    Returns:
        float: the angle
    """
    S = cross(V1,V2)
    s = S.norm()
    c = dot(V1,V2)
    return sign0(dot(S,N)) * math.atan2(s, c)

def signed_angle_3pts(A:Vec, B:Vec, C:Vec, N:Vec) -> float:
    """Signed angle between three points ABC with orientation givne by normal N

    Args:
        A (Vec): first point
        B (Vec): central point
        C (Vec): second point
        N (Vec): reference normal direction

    Returns:
        float: the angle
    """
    return signed_angle_2vec3D(A-B, C-B, N)

def angle_2vec2D(V1:Vec, V2:Vec) -> float:
    """Angle between two **2D** vectors

    Args:
        V1 (Vec): first vector
        V2 (Vec): second vector

    Returns:
        float: the angle
    """
    return math.atan2(V2.y, V2.x) - math.atan2(V1.y, V1.x)

def angle_2vec3D(V1:Vec, V2:Vec) -> float:
    """Angle between two **3D** vectors

    Args:
        V1 (Vec): first vector
        V2 (Vec): second vector

    Returns:
        float: the angle
    """
    C = cross(V1,V2)
    return math.atan2(C.norm(), dot(V1,V2))

def face_basis(*f):
    """
    Orthonormal basis of face
    Given three points A,B,C, returns a basis such that the first vector is along direction AB and third vector is normal to the plane ABC

    Args:
        *f: points of the face, either on a list or not

    Returns:
        Vec,Vec,Vec: an orthonormal 3D basis of the face
    """
    if len(f)==1: f = f[0]
    pA,pB,pC = (x for x in f)
    X = Vec.normalized(pB-pA)
    Z = Vec.normalized(cross(X, pC-pA))
    Y = Vec.normalized(cross(Z,X))
    return X,Y,Z

def triangle_area(A: Vec, B: Vec, C: Vec) -> float:
    """Area of 3D triangle ABC

    Args:
        A (Vec): first point
        B (Vec): second point
        C (Vec): third point

    Returns:
        float: area
    """
    return cross(B-A,C-A).norm()/2

def triangle_area_2D(A: Vec, B: Vec, C: Vec) -> float:
    """Area of 2D triangle ABC

    Args:
        A (Vec): first point
        B (Vec): second point
        C (Vec): third point

    Returns:
        float: area
    """
    return abs(det_2x2(B-A,C-A)/2)

def quad_area(A:Vec, B:Vec, C:Vec, D:Vec) -> float:
    """Area of the quad ABCD

    Args:
        A (Vec): first point
        B (Vec): second point
        C (Vec): third point
        D (Vec): fourth point

    Returns:
        float: area
    """
    return (triangle_area(A,B,C) + triangle_area(A,C,D) + triangle_area(B,C,D) + triangle_area(B,D,A))/2

def det_2x2(A:Union[complex,np.ndarray], B:Union[complex,np.ndarray]) -> float:
    """Computes a 2x2 determinant

    Parameters:
        A (Union[complex,np.ndarray]): first column vector. Can also be a complex number
        B (Union[complex,np.ndarray]): second column vector. Can also be a complex number

    Returns:
        float: the determinant
    """
    if isinstance(A,complex):
        ax,ay = A.real, A.imag
    else:
        ax,ay = A[0], A[1]
    if isinstance(B,complex):
        bx,by = B.real, B.imag
    else:
        bx,by = B[0], B[1]
    return ax*by - ay*bx

def det_3x3(*args) -> float:
    """Computes a 3x3 using the rule of Sarrus
    
    Parameters:
        args: Either a 3x3 numpy array representing a matrix, or 3 3x1 numpy array representing three column vectors

    Returns:
        float: the determinant
    """
    if len(args)==1:
        mat = args[0]
        assert mat.shape == (3,3)
    elif len(args)==3:
        A,B,C = args[0], args[1], args[2]
        mat = np.array([A,B,C])
    d = mat[0,0] * mat[1,1] * mat[2,2] + \
        mat[0,1] * mat[1,2] * mat[2,0] + \
        mat[0,2] * mat[1,0] * mat[2,1] - \
        mat[0,0] * mat[1,2] * mat[2,1] - \
        mat[0,1] * mat[1,0] * mat[2,2] - \
        mat[0,2] * mat[1,1] * mat[2,0]
    return d

def intersect_2lines2D(p1 : Vec, d1 : Vec, p2: Vec, d2 : Vec) -> Vec:
    """Computes the intersection of two lines in the plane

    Args:
        p1 (Vec): point on the first line
        d1 (Vec): direction vector of the first line
        p2 (Vec): point on the second line
        d2 (Vec): direction vector of the second line

    Returns:
        Vec: the intersection point. None if lines are parallel
    """
    p1,d1,p2,d2 = (u[:2] for u in (p1,d1,p2,d2))
    if abs(det_2x2(d1,d2))<1e-12 : return None #parallel lines
    n2 = Vec(d2.y, -d2.x)
    t = dot(p2-p1,n2)/dot(d1,n2)
    return p1+t*d1

def circumcenter(v1 : Vec, v2 : Vec, v3: Vec) -> Vec:
    """Circumcenter of the triangle formed by three points in the plane

    Parameters:
        v1 (Vec): first point
        v2 (Vec): second point
        v3 (Vec): third point

    Warning:
        Circumcenter of triangle (v1,v2,v3) may not lay inside the triangle

    Returns:
        Vec: coordinates of the circumcenter
    """
    p1 = (v1+v2)/2
    p2 = (v1+v3)/2
    d1 = v2-v1
    d2 = v3-v1
    d1 = Vec(d1[1], -d1[0])
    d2 = Vec(d2[1], -d2[0])
    S = intersect_2lines2D(p1, d1, p2, d2)
    return Vec(S.x, S.y, 0.)

def aspect_ratio(A : Vec, B : Vec, C : Vec) -> float:
    """
    Computes the aspect ratio of triangle ABC, defined as the ratio between the circumradius to twice the inradius.
    This ratio equals 1 for equilateral triangle and goes to zero as the triangle gets close to degenerate.

    Args:
        A (Vec): first point of the triangle
        B (Vec): second point of the triangle
        C (Vec): third point of the triangle

    Returns:
        float: the aspect ratio of triangle ABC

    Note:
        [https://stackoverflow.com/a/10290011](https://stackoverflow.com/a/10290011)
    """
    ab = distance(A,B)
    bc = distance(B,C)
    ca = distance(C,A)
    s = (ab+bc+ca)/2
    return ab*bc*ca/(8*(s-ab)*(s-bc)*(s-ca))


def distance_to_segment2D(P : Vec, A : Vec, B : Vec) -> float:
    """
    Computes, **in the plane**, the distance of point P to the segment [A;B]

    Args:
        P (Vec): Query point in 2D
        A (Vec): First segment extremity
        B (Vec): Second segment extremity

    Returns:
        float: the euclidean distance from P to [A;B]
    """
    P,A,B = P[:2], A[:2], B[:2]
    seg = B-A
    seg_length_sq = dot(seg,seg)
    if seg_length_sq<1e-12: 
        # segment is a single point
        return distance(P,A)
    t = max(0, min(1, dot(P-A, seg)/seg_length_sq))
    proj = A + t*seg
    return distance(P, proj)


def project_to_plane(P : Vec, N : Vec, orig : Vec) -> Vec:
    """Projects vector P onto the plane of normal N and passing through point 'orig'

    Args:
        P (Vec): query position to project
        N (Vec): normal vector of the plane

    Returns:
        Vec: P projected onto the plane
    """
    return P - dot(P-orig,N)/dot(N,N)*N


import numpy as np
from ..mesh.datatypes import *
from ..mesh.mesh_attributes import Attribute
from .misc_corners import corner_angles
from .misc_faces import face_area

@allowed_mesh_types(SurfaceMesh)
def scatter_vertices_to_corners(
    mesh : Mesh,
    vattr : Attribute,
    cattr : Attribute
) -> Attribute :
    """
    Given an attribute on vertices, distributes its values onto corresponding corners.

    Args:
        mesh (Mesh): the input mesh
        vattr (Attribute): input vertex attribute
        cattr (Attribute): output face corner attribute

    Returns:
        Attribute: cattr
    """
    for c,v in enumerate(mesh.face_corners):
        cattr[c] = vattr[v]
    return cattr

@forbidden_mesh_types(PointCloud,PolyLine)
def interpolate_vertices_to_faces(
    mesh : Mesh,
    vattr : Attribute, 
    fattr : Attribute
) -> Attribute:
    
    """Given an attribute on vertices, interpolates its value onto faces

    Parameters:
        mesh (Union[SurfaceMesh, VolumeMesh]): the input mesh
        vattr (Attribute): input vertex attribute 
        fattr (Attribute): output face attribute
    Returns:
        Attribute: fattr
    """
    fattr.clear()
    for f,F in enumerate(mesh.faces):
        for v in F:
            fattr[f] = fattr[f] + vattr[v]
    for f,F in enumerate(mesh.faces):
        fattr[f] /= len(F)
    return fattr

@allowed_mesh_types(SurfaceMesh)
def interpolate_faces_to_vertices(
    mesh : SurfaceMesh,
    fattr : Attribute, 
    vattr : Attribute,
    weight : str = "uniform")-> Attribute:
    """Given an attribute on vertices, interpolates its value onto faces

    Parameters:
        mesh (SurfaceMesh): the mesh
        fattr (Attribute): input face attribute
        vattr (Attribute): output face attribute
        weight (str): the way attributes are weighted in the sum.
            three possibilities :

            - uniform: every face will have weight 1
            
            - area: face have a weight proportionnal to their area
            
            - angle: face contribute to a vertex depending on the interior angle at this vertex
    Returns:
        Attribute: modified vattr

    Raises:
        Exception: fails if 'weight' is not in {'uniform', 'area', 'angle'}
    """
    weight = weight.lower()

    if weight not in {'uniform', 'area', 'angle'}:
        raise Exception(f"interpolate_vertices_to_faces: weight argument '{weight}' not recognized.\nPossibilites are {'uniform', 'area', 'angle'}.")

    if weight == "uniform":
        for v in mesh.id_vertices:
            for f in mesh.connectivity.vertex_to_faces(v):
                vattr[v] = vattr[v] + fattr[f]
        for v in mesh.id_vertices:
            vattr[v] /= len(mesh.connectivity.vertex_to_faces(v))
    
    elif weight == "area":
        if mesh.faces.has_attribute("area"):
            area = mesh.faces.get_attribute("area")
        else:
            area = face_area(mesh, persistent=False)
        total_area = np.zeros(len(mesh.vertices))
        for v in mesh.id_vertices:
            total_area[v] = 0.
            for f in mesh.connectivity.vertex_to_faces(v):
                vattr[v] = vattr[v] + fattr[f] * area[f]
                total_area[v] = total_area[v] + area[f]
        for v in mesh.id_vertices:
            vattr[v] /= total_area[v]

    elif weight == "angle":
        if mesh.face_corners.has_attribute("angles"):
            angles = mesh.face_corners.get_attribute("angles")
        else:
            angles = corner_angles(mesh, persistent=False)
        defects = np.zeros(len(mesh.vertices))
        for c, v in enumerate(mesh.face_corners):
            f = mesh.face_corners.adj(c)
            vattr[v] = vattr[v] + fattr[f] * angles[c]
            defects[v] = defects[v] + angles[c]
        for v in mesh.id_vertices:
            vattr[v] /= defects[v]
    return vattr


@allowed_mesh_types(SurfaceMesh)
def average_corners_to_vertices(
    mesh : SurfaceMesh,
    cattr : Attribute, 
    vattr : Attribute,
    weight : str = "uniform")-> Attribute:
    """Given an attribute on corners, computes a vertex attribute that is the average per vertex

    Parameters:
        mesh (SurfaceMesh): the mesh
        cattr (Attribute): input corner attribute
        vattr (Attribute): output face attribute
        weight (str): the way attributes are weighted in the sum.
            two possibilities:

            - uniform: every face will have weight 1
            - angle: corners contribute to a vertex depending on the interior angle at this vertex
    Returns:
        Attribute: modified vattr

    Raises:
        Exception: fails if 'weight' is not in {'uniform', 'angle'}
    """
    weight = weight.lower()

    if weight not in {'uniform', 'angle'}:
        raise Exception(f"average_corners_to_vertices: weight argument '{weight}' not recognized.\nPossibilites are {'uniform', 'angle'}.")

    if weight == "uniform":
        count = np.zeros(len(mesh.vertices))
        for c,v in enumerate(mesh.face_corners):
            vattr[v] = vattr[v] + cattr[c]
            count[v] += 1
        for v in mesh.id_vertices:
            vattr[v] /= count[v]

    elif weight == "angle":
        if mesh.face_corners.has_attribute("angles"):
            angles = mesh.face_corners.get_attribute("angles")
        else:
            angles = corner_angles(mesh, persistent=False)
        defects = np.zeros(len(mesh.vertices))
        for c, v in enumerate(mesh.face_corners):
            vattr[v] = vattr[v] + angles[c]*cattr[c]
            defects[v] += angles[c]
        for v in mesh.id_vertices:
            vattr[v] /= defects[v]

    return vattr


# @allowed_mesh_types(VolumeMesh)
# def interpolate_vertices_to_cells(mesh : Mesh, vattr : Attribute, cattr : Attribute)-> Attribute:
#     """
#     Given an attribute on vertices, interpolates its value onto cells

#     Parameters:
#         mesh (VolumeMesh): the input mesh
#         vattr (Attribute): input vertex attribute
#         cattr (Attribute): output cell attribute

#     Raises:
#         NotImplementedError: TODO
#     """
#     raise NotImplementedError


from .datatypes import *
from .mesh_data import RawMeshData
from .mesh import _instanciate_raw_mesh_data
from ..geometry import Vec
from ..utils import keyify, Logger

### Polyline Subdivision ###

@allowed_mesh_types(PolyLine)
def split_edge(polyline : PolyLine, edge_ind: int) -> PolyLine:
    """Split an edge of a polyline in half by adding a new vertex

    Note:
        Polylines do not need a specific subdivision class to be used in a `with` block since their connectivity is simpler. 
        Connectivity is cleared after processing.

    Args:
        mesh (PolyLine): the input polyline
        edge_ind (int): index of the edge to split

    Returns:
        PolyLine: the processed Polyline
    """
    A,B = polyline.edges[edge_ind]
    C = len(polyline.vertices)
    pC = (polyline.vertices[A]+polyline.vertices[B])/2
    polyline.vertices.append(pC)
    polyline.edges[edge_ind] += keyify(A,C)
    polyline.edges.append(keyify(B,C))
    polyline.connectivity.clear()
    return polyline


### Surface Subdivision ###

class SurfaceSubdivision(Logger):

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, verbose:bool = False):
        super().__init__("SurfaceSubdivision", verbose)
        self.mesh = mesh

    def __enter__(self):
        self.mesh = RawMeshData(self.mesh)
        self.mesh.face_corners.clear()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.mesh.prepare()
        self.mesh = _instanciate_raw_mesh_data(self.mesh, 2)

    @allowed_mesh_types(SurfaceMesh)
    def triangulate_face(self, face_id: int) :
        """Triangulates the face "face_id"

        Parameters:
            face_id (int): the face to triangulate
        """
        F = self.mesh.faces[face_id]
        if len(F)<4: return # nothing to do
        elif len(F)==4:
            A,B,C,D = self.mesh.faces[face_id]
            self.mesh.faces[face_id] = [A,B,D]
            self.mesh.faces.append([B,C,D])
        else:
            self.split_face_as_fan(face_id)

    def split_face_as_fan(self,  face_id: int) :
        """
        Adds a vertex at the barycenter of face 'face_id' and create a fan of triangles that replaces the face

        Parameters:
            face_id (int): the face to split
        """
        
        f = self.mesh.faces[face_id]
        pV = sum([Vec(self.mesh.vertices[a]) for a in f ])/len(f) # barycenter
        iV = len(self.mesh.vertices)
        self.mesh.vertices.append(pV)
        self.mesh.faces[face_id] = [f[0], f[1], iV]
        nf=len(f)
        for k in range(1, nf):
            self.mesh.faces.append([f[k], f[(k+1)%nf], iV])
        for v in f:
            self.mesh.edges.append(keyify(v,iV))

    def triangulate(self):
        """Triangulates all faces of a mesh.
            Calls triangulate_face on every faces.
        """       
        for f in self.mesh.id_faces:
            if len(self.mesh.faces[f])!= 3 :
                self.triangulate_face(f)

    def loop_subdivision(self, n: int = 1):
        """Subdivides triangles of a mesh in 4 triangles by splitting along middle of edges.
            If the mesh is not triangulated, will triangulate the mesh first.

        References:
            [https://en.wikipedia.org/wiki/Loop_subdivision_surface](https://en.wikipedia.org/wiki/Loop_subdivision_surface)
        
        Parameters:
            n (int, optional): Number of times the subdivision is applied. Defaults to 1.
        """
        self.triangulate()

        for _ in range(n):
            newMeshData = RawMeshData()
            newMeshData.vertices += self.mesh.vertices
            # cut every edge in half
            half = dict()
            for (A,B) in self.mesh.edges:
                C = len(newMeshData.vertices)
                pC = (self.mesh.vertices[A] + self.mesh.vertices[B])/2
                newMeshData.vertices.append(pC)
                half[keyify(A,B)]=C
            new_edges = set()
            for f in self.mesh.id_faces:
                A,B,C = self.mesh.faces[f]
                mAB = half[keyify(A,B)]
                mBC = half[keyify(B,C)]
                mCA = half[keyify(C,A)]
                for new_tri in [
                    (mAB,mBC,mCA),
                    (A,mAB,mCA),
                    (B,mBC,mAB),
                    (C,mCA,mBC)
                ]:
                    newMeshData.faces.append(new_tri)
                for new_edge in [
                    (A,mAB),(mAB,B),(B,mBC),(mBC,C),(C,mCA),(mCA,A), (mAB,mBC),(mBC,mCA),(mCA,mAB)
                ]:
                    new_edges.add(keyify(new_edge))
            
            newMeshData.edges += list(new_edges)
            self.mesh = newMeshData

    @allowed_mesh_types(SurfaceMesh)
    def subdivide_triangles_6(self, repeat: int = 1) -> SurfaceMesh:
        """Subdivides triangles of a mesh in 6 triangles by adding a point at the barycenter and three middles of edges.
            If the mesh is not triangulated, will triangulate the mesh first.
            
            Same operation as subdividing into 3 quads and then splitting the quads along the corner-barycenter diagonal.

        Parameters:
            repeat (int, optional): number of successive subdivisions. Eventual first triangulation does not count. Defaults to 1.
        """
        for _ in range(repeat):
            self.subdivide_triangles_3quads()
            self.triangulate()

    @allowed_mesh_types(SurfaceMesh)
    def subdivide_triangles_3quads(self) -> SurfaceMesh:
        """Subdivides triangles of a mesh in 3 quads by adding a point at the barycenter and three middles of edges.
            If the mesh is not triangulated, will triangulate the mesh first.
        """
        self.triangulate()
        newMeshData = RawMeshData()
        newMeshData.vertices += self.mesh.vertices
        # cut every edge in half
        half = dict()
        for e in self.mesh.id_edges:
            A,B = self.mesh.edges[e]
            C = len(newMeshData.vertices)
            pC = (self.mesh.vertices[A] + self.mesh.vertices[B])/2
            newMeshData.vertices.append(pC)
            half[keyify(A,B)]=C

        bary = dict()
        for iF,F in enumerate(self.mesh.faces):
            pS = sum([self.mesh.vertices[u] for u in F])/3
            bary[iF] = len(newMeshData.vertices)
            newMeshData.vertices.append(pS)

        for f in self.mesh.id_faces:
            A,B,C = self.mesh.faces[f]
            mAB = half[keyify(A,B)]
            mBC = half[keyify(B,C)]
            mCA = half[keyify(C,A)]
            S = bary[f]
            
            for new_face in [
                [A, mAB, S, mCA],
                [B, mBC, S, mAB],
                [C, mCA, S, mBC],
            ]:
                newMeshData.faces.append(new_face)
        self.mesh = newMeshData

@allowed_mesh_types(SurfaceMesh)
def split_double_boundary_edges_triangles(mesh : SurfaceMesh) -> SurfaceMesh:
    """
    A triangle with double edge on the boundary can occur on the border in a case like :

    ```
        /\\ 
    ___/__\\___
    ```

    This function detects every occurrences of such a configuration and split the problematic
    triangle in three by adding a new vertex in the middle. Uses the SurfaceSubdivision class.

    Parameters:
        mesh (Mesh): the mesh (modified in place)

    Raises:
        Exception: Isolated vertex
        Raised when a vertex of the mesh has degree < 2, which mean that the mesh is not manifold

    Returns:
        SurfaceMesh: the modified input mesh
    """
        
    deg = [0]*len(mesh.vertices)
    for (a,b) in mesh.edges:
        deg[a] += 1
        deg[b] += 1

    pb_faces = []
    for i,f in enumerate(mesh.faces):
        for v in f:
            if deg[v]<2: 
                raise Exception("Isolated vertex")
            if deg[v]==2:
                pb_faces.append(i)
                break
    if pb_faces:
        with SurfaceSubdivision(mesh) as subdv:
            for f in pb_faces: # Triangulate face with a vertex in the middle
                subdv.split_face_as_fan(f)
    return mesh

### Volume Subdivision ###

class VolumeSubdivision(Logger):

    @allowed_mesh_types(VolumeMesh)
    def __init__(self, mesh : VolumeMesh, verbose:bool=False):
        super().__init__("VolumeSubdivision", verbose=verbose)
        self.mesh = mesh
        self.conn = None # connectivity

    def __enter__(self):
        self.conn = self.mesh.connectivity
        self.conn._compute_cell_adj()
        self.mesh = RawMeshData(self.mesh)
        self.mesh.face_corners.clear()
        self.mesh.cell_corners.clear()
        self.mesh.cell_faces.clear()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.mesh.prepare()
        self.mesh = _instanciate_raw_mesh_data(self.mesh, 3)

    def split_cell_as_fan(self, cell_id:int):
        """
        Adds a vertex at the barycenter of cell 'cell_id' and create a fan of tetrahedra that replaces the cell.
        If the cell 'cell_id' is not a tetrahedron, does nothing.

        Parameters:
            mesh (VolumeMesh): the input mesh
            cell_id (int): the (tetrahedral) cell to split

        Returns:
            (VolumeMesh): the modified input mesh
        """
        if len(self.mesh.cells[cell_id]) != 4 : return # cell is not a tet
        
        A,B,C,D = self.mesh.cells[cell_id] 
        pA,pB,pC,pD = (self.mesh.vertices[_v] for _v in (A,B,C,D))
        bary = 0.25*(pA+pB+pC+pD)
        ibary = len(self.mesh.vertices)
        self.mesh.vertices.append(bary)
        self.mesh.cells[cell_id] = (ibary,B,C,D)
        self.mesh.cells += [
            (A,ibary,C,D),
            (A,B,ibary,D),
            (A,B,C,ibary)
        ]

    def split_tet_from_face_center(self, face_id : int):
        """Split the triangle `face_id` into three triangles by adding a point at its barycenter,
        and split adjacent tetrahedra accordingly.

        Args:
            face_id (int): index of the face to split
        """
        f = self.mesh.faces[face_id]
        if len(f) != 3 : return
        A,B,C = f
        icenter = len(self.mesh.vertices)
        pcenter = sum([Vec(self.mesh.vertices[a]) for a in f ])/3 # barycenter
        self.mesh.vertices.append(pcenter)
        
        for c in self.conn.face_to_cells(face_id):
            iF = self.conn.in_cell_face_index(c,face_id)
            new_cells = []
            for i in range(4):
                if i==iF : continue # opposite point in tet from face
                cell = [_x for _x in self.mesh.cells[c]]
                cell[i] = icenter
                new_cells.append(cell)
            self.mesh.cells[c] = new_cells[0]
            self.mesh.cells.append(new_cells[1])
            self.mesh.cells.append(new_cells[2])
        
        self.mesh.faces[face_id] = [A,B,icenter]
        self.mesh.faces.append([icenter, B, C])
        self.mesh.faces.append([A, icenter, C])


import numpy as np
from dataclasses import dataclass
from enum import Enum
from collections import deque
from ..utils import check_argument, PriorityQueue
from ..geometry import AABB, distance, Vec

class KDTree:

    @dataclass
    class Node:
        id : int # id of the node
        split_axis : int # which axis does the node split
        parent : int  # id of parent node (-1 if root node)
        split_value : float = None
        left : int = None # id of the left child
        right : int = None # id of the right child
        bb : AABB = None # bounding box

    @dataclass
    class Leaf:
        id : int # id of the leaf
        split_axis : int = None # which axis does the leaf split (useful at build time, not so much after)
        parent : int = None # id of the parent node
        points : np.ndarray = None # indices of contained points
        bb : AABB = None # bounding box of contained points

        @property
        def size(self):
            return self.points.size

    class BuildStrategy(Enum):
        BALANCED = 0 # Consider the median value as the pivot at each depth. Slower but garanteed to provide a balanced tree
        FAST = 1 # Take the median of a random sampling of 100 elements. Faster than fully balanced, almost balanced in practice
        RANDOM = 2 # Consider a random pivot at each depth

        @classmethod
        def from_string(cls, txt : str):
            if txt.lower() == "balanced": return cls.BALANCED
            if txt.lower() == "fast": return cls.FAST
            if txt.lower() == "random": return cls.RANDOM

    def __init__(self, points : np.ndarray, max_leaf_size : int=10, strategy: str="fast"):
        """Implementation of a kd-tree for space partitionning and fast nearest point queries.

        Args:
            points (np.ndarray): array of points (shape (N,d)) to consider. Can be of any dimension d, though the kd-tree algorithm have poor performance when the dimension grows.
            max_leaf_size (int, optional): Maximum number of points inside a leaf. Defaults to 10.
            strategy (str, optional): Node splitting strategy. Can be either 'fast', 'balanced' or 'random'. Defaults to "fast".

        Raises:
            Exception: fails if the points are not of the correct (N,d) shape.
        """
        
        points = np.array(points)
        if len(points.shape)!=2:
            raise Exception("Expected an array of points of shape (N,dim)")
        self.points = points
        self.n_pts, self.dim = points.shape
        
        check_argument("strategy", strategy.lower(), str, ["balanced", "fast", "random"])
        self.build_strategy = KDTree.BuildStrategy.from_string(strategy)

        self.nodes = []
        self._nid = 0
        root = self._new_leaf(0, None, np.arange(self.n_pts)) # initialize a leaf node that contains all points
        root.bb = AABB.infinite(self.dim)
        queue = deque()
        queue.append(root)
        while len(queue)>0:
            leaf = queue.popleft()
            if leaf.size <= max_leaf_size: # the leaf has the correct size -> add it to the tree
                self.nodes.append(leaf)
            else: # the leaf needs to be split
                # split the points according to the current axis
                split_value, pts_less, pts_more = self._split_points(leaf.points, leaf.split_axis)
                
                # we create a new node to replace the original leaf and append two leaves that will be its children
                node = KDTree.Node(leaf.id, leaf.split_axis, parent=leaf.parent, bb=leaf.bb, split_value=split_value)
                leaf_less = self._new_leaf((leaf.split_axis + 1)%self.dim, leaf.id, pts_less)
                leaf_more = self._new_leaf((leaf.split_axis + 1)%self.dim, leaf.id, pts_more)
                node.left, node.right = leaf_less.id, leaf_more.id

                # split the bounding box
                bbmax_less = np.copy(leaf.bb.maxi)
                bbmax_less[leaf.split_axis] = split_value
                bbmin_more = np.copy(leaf.bb.mini)
                bbmin_more[leaf.split_axis] = split_value
                leaf_less.bb = AABB(leaf.bb.mini, bbmax_less)
                leaf_more.bb = AABB(bbmin_more, leaf.bb.maxi)

                # append and continue
                self.nodes.append(node)
                queue.append(leaf_less)
                queue.append(leaf_more)

    def _new_leaf(self, axis, parent, points):
        leaf = KDTree.Leaf(self._nid, axis, parent, points)
        self._nid +=1
        return leaf

    def _split_points(self, pt_idx, axis):
        pts_ax = self.points[pt_idx,axis] # 1D array of the considered coordinate to split 
        pivot = self._find_pivot(pts_ax)
        pivot_filter = pts_ax <= pivot
        idx_less = np.extract(pivot_filter, pt_idx)
        idx_more = np.extract(~pivot_filter, pt_idx)
        return pivot, idx_less, idx_more
    
    def _find_pivot(self, pts_ax):
        if self.build_strategy == KDTree.BuildStrategy.FAST:
            samples = np.random.choice(pts_ax, min(50, pts_ax.size), replace=False)
            return np.median(samples)
        elif self.build_strategy == KDTree.BuildStrategy.BALANCED:
            return np.median(pts_ax)
        elif self.build_strategy == KDTree.BuildStrategy.RANDOM:
            return np.random.choice(pts_ax, 1)[0]
        else:
            raise Exception("Should not happen.")

    def is_leaf(self, node_id:int) -> bool:
        return isinstance(self.nodes[node_id], KDTree.Leaf)

    def query(self, pt:Vec, k:int=1) -> list:
        """Query the tree to find the k nearest neighbors of point 'pt'

        Args:
            pt (Vec): the position to query
            k (int, optional): number of nearest points to query. Defaults to 1.

        Returns:
            list: the indices of the k nearest points of 'pt' in increasing distance order
        """
        found = PriorityQueue()
        n_found = 0
        queue = deque()
        queue.append(0) # start from the root
        while len(queue)>0:
            node_id = queue.pop() # stack order for depth first search
            if self.is_leaf(node_id):
                leaf = self.nodes[node_id]
                for idx in leaf.points:
                    found.push(idx, -distance(self.points[idx], pt))
                    n_found += 1
                    while n_found>k:
                        # keep only k closest
                        found.pop() 
                        n_found -= 1
            else:
                node = self.nodes[node_id]
                furthest_so_far = -found.front.priority if not found.empty() else float("inf")
                dist_left = self.nodes[node.left].bb.distance(pt)
                dist_right = self.nodes[node.right].bb.distance(pt)                
                for dist,child in sorted([(dist_left, node.left), (dist_right,node.right)]):
                    if furthest_so_far > dist:
                        queue.append(child)
        return [found.pop().x for _ in range(n_found)][::-1]
    
    def query_radius(self, pt:Vec, r:float) -> list:
        """Query the tree to find all points that are at distance <= r from the position 'pt'.

        Args:
            pt (Vec): the position to query
            r (float): radius of the query ball

        Returns:
            list: all indices of points that are at distance at most r from the query point.
        """
        queue = deque()
        found_pt = []
        queue.append(0)
        while len(queue)>0:
            node_id = queue.popleft()
            if self.nodes[node_id].bb.distance(pt)>r : 
                continue # no point inside the node that have distance < r
            if self.is_leaf(node_id): # read leaf
                leaf = self.nodes[node_id]
                found_pt += [idx for idx in leaf.points if distance(self.points[idx],pt)<=r]
            else: # read node
                node = self.nodes[node_id]
                queue.append(node.left)
                queue.append(node.right)
        return found_pt  



from ..mesh.datatypes import *
from ..geometry import Vec
from .misc_faces import face_area
from .misc_cells import cell_volume

@allowed_mesh_types(SurfaceMesh)
def euler_characteristic(mesh : SurfaceMesh) -> int:
    """
    Computes the Euler characteristic of a surface mesh, given as V-E+F
    
    See https://en.wikipedia.org/wiki/Euler_characteristic

    Args:
        mesh (SurfaceMesh): the mesh

    Returns:
        (int): the Euler characteristic
    """
    
    v = len(mesh.vertices)
    e = len(mesh.edges)
    f = len(mesh.faces)
    return v-e+f

# @allowed_mesh_types(SurfaceMesh)
# def genus(mesh : SurfaceMesh) -> int:
#     """Computes the genus g of a mesh.
#     The genus is the number of "holes" a surface contains. It is linked to the euler characteristic X by
#     X = 2 - 2g

#     Parameters:
#         mesh (Mesh): the mesh

#     Returns:
#         (int): the Euler characteristic
#     """
#     X = euler_characteristic(mesh)
#     return 1 - X//2

@forbidden_mesh_types(PointCloud)
def mean_edge_length(mesh : Mesh, n : int = None) -> float:
    """
    Estimation of mean edge length

    Parameters:
        mesh (Mesh): input mesh
        n (int, optional): 
            Early stopping for number of edges to consider in mean computation. 
            If set to None, considers all the edges.
            Defaults to None.

    Returns:
        float: the computed mean edge length
    """
    l = 0
    if n is None: n = len(mesh.edges)
    for k in range(min(n, len(mesh.edges))):
        a,b = (Vec(mesh.vertices[u]) for u in mesh.edges[k])
        l += (b-a).norm()
    return l/n

@allowed_mesh_types(SurfaceMesh,VolumeMesh)
def mean_face_area(mesh : SurfaceMesh, n : int = None) -> float:
    """Estimation of mean face area

    Parameters:
        mesh (Mesh): input mesh
        n (int, optional): 
            Early stopping for number of faces to consider in mean computation. 
            If set to None, considers all the faces.
            Defaults to None.

    Returns:
        float: the computed mean face area
    """
    if mesh.faces.has_attribute("area"):
        farea = mesh.faces.get_attribute("area")
    else:
        farea = face_area(mesh)
    if n is None: n = len(mesh.faces)
    res = 0
    for k in range(min(n, len(mesh.faces))):
        res += farea[k]
    return res/n

@allowed_mesh_types(VolumeMesh)
def mean_cell_volume(mesh : VolumeMesh, n : int = None) -> float:
    """Estimation of mean cell volume

    Parameters:
        mesh (VolumeMesh): input mesh
        n (int, optional): 
            Early stopping for number of cells to consider in mean computation. 
            If set to None, considers all the cells.
            Defaults to None.

    Returns:
        float: the computed mean cell volume
    """
    if mesh.cells.has_attribute("volume"):
        cvol = mesh.cells.get_attribute("volume")
    else:
        cvol = cell_volume(mesh)
    if n is None: n = len(mesh.cells)
    res = 0
    for k in range(min(n, len(mesh.cells))):
        res += cvol[k]
    return res/n

@allowed_mesh_types(SurfaceMesh)
def total_area(mesh : SurfaceMesh) -> float:
    """Sum of face areas

    Parameters:
        mesh (Mesh): input mesh

    Returns:
        float: the computed mean face area
    """
    if mesh.faces.has_attribute("area"):
        farea = mesh.faces.get_attribute("area")
    else:
        farea = face_area(mesh, persistent=False)
    return sum([farea[iF] for iF in mesh.id_faces])

def barycenter(mesh : Mesh) -> Vec:
    """Barycenter of the vertices of the mesh

    Args:
        mesh (Mesh): input mesh

    Returns:
        Vec: coordinates of the barycenter
    """
    return sum(mesh.vertices)/len(mesh.vertices)


from ..mesh.mesh_attributes import Attribute, Attribute, ArrayAttribute
from ..mesh.datatypes import *
from .. import geometry as geom
import numpy as np

@allowed_mesh_types(SurfaceMesh)
def corner_angles(mesh : SurfaceMesh, name = "angles", persistent:bool=True, dense:bool = True) -> Attribute:
    """Attribute storing the angles of a face at a vertex

    Parameters:
        mesh (SurfaceMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "angles".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute: one float per face corner
    """
    if persistent:
        angles = mesh.face_corners.create_attribute(name, float, dense=True)
    else:
        angles = ArrayAttribute(float, len(mesh.face_corners)) if dense else Attribute(float)
    c = 0
    for face in mesh.faces:
        n = len(face)
        for i in range(n):
            iPrev,iV,iNext = face[(i-1)%n], face[i], face[(i+1)%n]
            pPrev, pV, pNext = mesh.vertices[iPrev], mesh.vertices[iV], mesh.vertices[iNext]
            angles[c] = geom.angle_3pts(pPrev, pV, pNext)
            c += 1
    return angles

@allowed_mesh_types(SurfaceMesh)
def cotangent(mesh : SurfaceMesh, name = "cotan", persistent=True, dense=True) -> Attribute:
    """Attribute storing the cotangents of each face at each vertex.
    
    Warning: 
        Only works if the mesh is triangulated (ie every faces are triangles)

    Parameters:
        mesh (SurfaceMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "cotan".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute(: one float per face corner

    Raises:
        Exception: fails if the mesh is not triangulated
    """

    if not mesh.is_triangular():
        raise Exception("Tried to compute cotangents on a non-triangulated mesh")

    if persistent:
        if mesh.face_corners.has_attribute(name):
            cot = mesh.face_corners.get_attribute(name)
        else:
            cot = mesh.face_corners.create_attribute(name, float, dense=True)
    else:
        cot = ArrayAttribute(float, len(mesh.face_corners)) if dense else Attribute(float)

    if mesh.face_corners.has_attribute("angles"):
        # Compute cotan from angles
        angles = mesh.face_corners.get_attribute("angles")
        for c in mesh.id_corners:
            cot[c] = -np.tan(angles[c] + np.pi/2)
    else:
        for i, (iA,iB,iC) in enumerate(mesh.faces):
            pA, pB, pC = (mesh.vertices[_i] for _i in (iA,iB,iC))
            cot[3*i] = geom.cotan(pC, pA, pB)
            cot[3*i+1] = geom.cotan(pA, pB, pC)
            cot[3*i+2] = geom.cotan(pB, pC, pA)
    return cot


import numpy as np
from osqp import OSQP
from dataclasses import dataclass
import scipy.sparse as sp
from types import FunctionType

from tqdm import tqdm, trange
from tqdm.utils import _term_move_up
prefix = _term_move_up() + '\r'

from ..utils import Logger, get_osqp_lin_solver
from .. import geometry as geom

@dataclass
class LMParameters:
    """
    Hyper parameters for the Levenberg-Marquardt algorithm
    """
    N_ITER_MAX    : int   = 1000 # stopping criterion on maximal number of iterations
    ENERGY_MIN    : float = 1e-7 # stopping criterion for reaching zero
    MIN_STEP_NORM : float = 1e-5 # stopping criterio, on ||x_n - x_{n-1}||
    MIN_GRAD_NORM : float = 1e-5 # stopping criterion on gradient projected on constraints
    MIN_DELTA_E   : float = 1e-5 # stopping criterion
    MU_MAX        : float = 1e8  # stopping criterion on mu going to +infty
    MU_MIN        : float = 1e-8 # minimal value for mu
    alpha         : float = 0.5  # if iteration is a success, mu = alpha * mu
    beta          : float = 2.   # if iteration fails, mu = beta * mu

@dataclass
class LMVerboseOptions:
    logger_verbose : bool = True  # main verbose for mouette logging
    solver_verbose : bool = False # prints outputs of OSQP solver
    use_tqdm       : bool = True  # progression bar using tqdm
    log_frequency  : int  = 1     # display an info message every 'log_frequency' iterations

@dataclass
class LMFunction:
    name : str # name of the function
    fun  : FunctionType # x -> f(x), J(x) where J is the jacobian
    fun_noJ : FunctionType = None # x -> f(x) without jacobian
    weight : float = 1. # multiplicative weight

    def call(self, X):
        F,J = self.fun(X)
        return self.weight*F, self.weight*J

    def call_noJ(self,X):
        return self.weight*self.fun_noJ(X)

class LevenbergMarquardt(Logger):
    """
    Levenberg-Marquardt algorithm for non-linear least-square minimization.

    References:
        [1] https://en.wikipedia.org/wiki/Levenberg%E2%80%93Marquardt_algorithm
        [2] Constrained Levenberg-Marquardt Method with global complexity bounds, Marumo et al.
    """

    def __init__(self,
        HP : LMParameters = LMParameters(), 
        verbose : LMVerboseOptions = LMVerboseOptions(),
        **kwargs):
        super().__init__("LMopt",verbose.logger_verbose)
        self.verbose_options = verbose
        self.HP = HP

        self.X = None # variables

        ### Linear Constraints
        self._cstM = None
        self._cstL = None
        self._cstU = None

        ### Metric matrix
        self._W : sp.csc_matrix = None

        ### Function to optimize
        self._functions = []

        ### Others
        self._stop_criterion_instance = None # OSQP instance to compute projected gradient norm
        
        ### Additionnal parameters
        self._linsys_solver = kwargs.get("lin_solver", get_osqp_lin_solver())

    def register_constraints(self, A : sp.spmatrix, l : np.ndarray = None, u : np.ndarray = None):
        """
        Adds linear constraints to the optimization problem :

        min_X ||F(X)||² s.t.  l <= AX <= u

        Args:
            cstMat (sp.spmatrix): constraint matrix, either dense or sparse
            cstL (np.ndarray, optional): vector l. Defaults to None.
            cstR (np.ndarray, optional): vector u. Defaults to None.
        """

        if isinstance(A, sp.spmatrix):
            self._cstM = A.tocsc()
        elif isinstance(A, np.ndarray):
            assert len(A.shape)==2 # we make sure A is indeed a matrix
            self._cstM = sp.csc_matrix(A)
        self._cstL = l
        self._cstU = u

    def register_function(self, fun, fun_noJ=None, weight:float=1., name:str=None):
        """
        Adds a function to minimize.

        Args:
            fun (python callable): A function taking a single argument X (np.array) returning a vector of values F and a (sparse or dense) jacobian matrix J

            fun_noJ (python callable, optional): The same function that avoids the computation of the jacobian for speed purposes. If not provided, will use the function provided above and ignore the jacobian.

            weight (float, optional): real weight to be applied to the function. Defaults to 1..
            
            name (str, optional): Name of the function in the logs. If not specified, the function will be given a default name.
        """
        if name is None:
            name = f"fun{len(self._functions)}"
        if fun_noJ is None:
            fun_noJ = lambda x : fun(x)[0]
        self._functions.append( LMFunction(name, fun, fun_noJ, weight) )

    def set_metric_matrix(self, W : sp.spmatrix):
        # TODO : some assertions on the matrix
        self._W = W.tocsc()

    def energy(self, X, jac=True, which=None):
        funs_to_call = self._functions if which is None else [fun for fun in self._functions if fun.name in which]
        if jac:
            F,J = [],[]
            for f in funs_to_call:
                fi,ji = f.call(X)
                F.append(fi)
                J.append(ji)
            return F,J
        return [f.call_noJ(X) for f in funs_to_call]

    def optimize(self, x_init: np.ndarray) -> float:
        """
        Alias of the 'run' method.

        Args:
            x_init (np.ndarray): initial values for the variables.

        Returns:
            float: final value of the energy
        """
        return self.run(x_init)

    def run(self, x_init : np.ndarray):
        """
        Runs the optimizer

        Args:
            x_init (np.ndarray): initial values for the variables.

        Returns:
            float: final value of the energy
        """

        self.X = x_init # variable vector
        
        if self.HP.N_ITER_MAX <= 0 :
            # return current energy
            en = self.energy(x_init, jac=False)
            return 0.5*sum(np.dot(e,e) for e in en)

        if not self._functions: 
            # No functions to optimize
            self.log("No function to optimize.")
            return 0.

        if self._W is None:
            # Metric matrix is Id by default (classical L2 norm)
            self._W = sp.identity(x_init.size, format="csc")

        mu = 1.
        mu_avg = mu
        update = True
        grad_norm = None
        step_norm = 0.
        RelDeltaE = 0.

        # iterable for optimization steps
        if self.verbose_options.logger_verbose and self.verbose_options.use_tqdm:
            print()
            iterobj = trange(self.HP.N_ITER_MAX, total=self.HP.N_ITER_MAX, position=1, leave=False, ncols=100, unit="")
        else:
            iterobj = range(self.HP.N_ITER_MAX)

        try:
            # Iteration loop
            for it in iterobj:                

                if update: # energy only changes if we performed a step last iteration
                    flist, Jlist = self.energy(self.X)
                    Jx = sp.vstack(Jlist).tocsr()
                    fx = np.concatenate(flist)
                    Jt = Jx.transpose()
                    JtJ = Jt.dot(Jx)
                    q = Jt.dot(fx) # gradient

                    Ex = np.dot(fx,fx)/2

                    if Ex<self.HP.ENERGY_MIN:
                        return self._end_optimization(Ex, "Ex < Ex_min") # zero found
                    if np.isnan(Ex):
                        return self._end_optimization(Ex, "NaN value in energy")
                    if np.isposinf(Ex):
                        mu = self.HP.beta*mu
                        update=False
                        continue
                    grad_stop, grad_norm = self._stop_criterion_projgrad(q)
                    if grad_stop:
                        return self._end_optimization(Ex, "projected grad norm on constraints < min_grad_norm")

                if mu>self.HP.MU_MAX:
                    return self._end_optimization(Ex, "mu > mu_max")
                    
                gamma = mu * np.sqrt(2*Ex)
                osqp_instance = OSQP()
                if self._cstM is not None:
                    # OSQP solve inside iteration loop computes the increment x_{n-1} - x_n, so constraints have to be shifted
                    xCst = self._cstM @ self.X
                    cstL = self._cstL - xCst
                    cstU = self._cstU - xCst
                else:
                    cstL,cstU = None, None
                osqp_instance.setup(JtJ + gamma*self._W, q=q, A=self._cstM, l=cstL, u=cstU,
                    verbose=self.verbose_options.solver_verbose,
                    eps_abs=1e-3, eps_rel=1e-3,
                    max_iter=100, polish=True, check_termination=10, 
                    adaptive_rho=True, linsys_solver= self._linsys_solver)
                s = osqp_instance.solve().x # computed increment

                if s[0] is not None:
                    ms = fx + Jx.dot(s)
                    ms = np.dot(ms,ms)/2 + gamma*np.dot(s,s)/2
                    fxs = self.energy(self.X + s, jac=False)
                    fxs = np.concatenate(fxs)
                    Exs = np.dot(fxs,fxs)/2

                update = (s is not None) and (Exs <= ms)  # whether to make a step (True) or increase mu (False)
                if update:
                    RelDeltaE = abs(Exs-Ex)/Ex
                    if RelDeltaE < self.HP.MIN_DELTA_E :
                        return self._end_optimization(Ex, "Relative progression of energy < ΔE_min")
                    self.X += s
                    step_norm = geom.norm(s)
                    if step_norm < self.HP.MIN_STEP_NORM :
                        return self._end_optimization(Ex, "Step norm < min_step_norm")
                    mu, mu_avg = max(self.HP.MU_MIN, self.HP.alpha*mu_avg), mu
                else:
                    mu = self.HP.beta*mu

                if self.verbose_options.logger_verbose and self.verbose_options.log_frequency >0 and it%self.verbose_options.log_frequency==0:
                    energies = [np.dot(_f,_f)/2 for _f in flist]
                    if self.verbose_options.use_tqdm:
                        tqdm_log = prefix
                        if len(self._functions)>1:
                            for n,e in zip([f.name for f in self._functions], energies):
                                tqdm_log+="{}: {:.2E} | ".format(n,e)
                            tqdm_log += "Total: {:.2E} | Grad: {:.2E} | ".format(Ex, grad_norm)
                        else:
                            tqdm_log += "{}: {:.2E} | Grad: {:.2E} | ".format(self._functions[0].name, Ex, grad_norm)
                        tqdm_log += "ΔE {:.2E} | Step {:.2E} | Mu: {:.2E}".format(RelDeltaE, step_norm, mu)
                        tqdm_log += " " * 10
                        tqdm.write(prefix)
                        tqdm.write(tqdm_log)
                    else:
                        log = f"{it+1}/{self.HP.N_ITER_MAX} | "
                        if len(self._functions)>1:
                            for n,e in zip([f.name for f in self._functions], energies):
                                log+="{}: {:.2E} | ".format(n,e)
                            log += "Total: {:.2E} | Grad: {:.2E} | ".format(Ex, grad_norm)
                        else:
                            log += "{}: {:.2E} | Grad: {:.2E} | ".format(self._functions[0].name, Ex, grad_norm)
                        log += "ΔE {:.2E} | Step {:.2E} | Mu: {:.2E}".format(RelDeltaE, step_norm, mu)
                        print(log)
                    
        except KeyboardInterrupt:
            self.log("Manual interruption")
        return self._end_optimization(Ex, "max iteration reached")

    def _stop_criterion_projgrad(self, G):
        # Projected gradient on the orthogonal of space spanned by constraints should have small norm
        if self._cstM is None:
            xnorm = np.sqrt(np.dot(G,G))
        else:
            n = self._cstM.shape[1]
            if self._stop_criterion_instance is None:
                self._stop_criterion_instance = OSQP()
                self._stop_criterion_instance.setup(
                    sp.eye(n,format=("csc")), q=-2*G, A = self._cstM, l = self._cstL, u = self._cstU,
                    verbose=self.verbose_options.solver_verbose)
            else:
                self._stop_criterion_instance.update(q=-2*G)
            x = self._stop_criterion_instance.solve().x
            if x is None: return False, 0.
            xnorm = np.sqrt(np.dot(x,x))
        return xnorm<self.HP.MIN_GRAD_NORM, xnorm

    def _end_optimization(self, energy, message):
        if self.verbose:
            if self.verbose_options.use_tqdm: print("\n")
            self.log("End of optimization: "+message)
            self.log(f"Final Energy: {energy:.4E}")
            if self.verbose_options.use_tqdm: print("\n\r")
        return energy        


from .utilities import *
from .unionfind import UnionFind
from .priority_queue import PriorityQueue
from .iterators import *
from . import maths
from .argument_check import *
from .osqp_lin_solve import get_osqp_lin_solver


from ..mesh.datatypes import *
from ..operators.laplacian_op import laplacian
from ..geometry import distance

import numpy as np
import scipy.sparse as sp

@allowed_mesh_types(SurfaceMesh)
def smooth(mesh : SurfaceMesh, n_iter=100, mode="cotan", keep_border=False, damping=0.0) -> SurfaceMesh:
    """
    Applies Laplacian smoothing to a mesh

    Parameters:
        mesh (Mesh): 
            The mesh to smooth. Operation is done in place
        
        n_iter (int, optional): 
            Number of steps. Defaults to 100.
        
        mode (str, optional): 
            if "cotan": uses cotan expression of Laplace-Beltrami operator. 
            if "connectivity": uses simple connectivity Laplacian.
            Defaults to "cotan".
        
        keep_border (bool, optional):
            Fix the boundary. Defaults to False.

        damping (float, optional): 
            Damping coefficient. Will consider matrix (L+ + damping*Id). 
            Defaults to 0.0.

    Returns:
        (SurfaceMesh): the smoothed input mesh
    """
    
    A = laplacian(mesh, cotan=(mode=="cotan"))

    if damping>0.:
        A = damping*A + (1-damping)*sp.eye(A.shape[0])
    V = np.array(mesh.vertices)
    dist_begin = distance(V[0], V[1])

    if keep_border:
        mask = np.array([not mesh.is_vertex_on_border(i) for i in mesh.id_vertices]).astype(bool)

    for _ in range(n_iter):
        if keep_border:
            V[mask] = A.dot(V)[mask]
        else:
            V = A.dot(V)

    dist_end = distance(V[0], V[1])
    vA =  V[0]
    V = (V-vA)*(dist_begin/dist_end) + vA

    for i in mesh.id_vertices:
        mesh.vertices[i] = V[i,:]
    return mesh


from ..worker import Worker

from ...mesh.datatypes import *
from ...mesh.mesh_attributes import ArrayAttribute
from ...mesh.mesh import copy

from ...geometry import Vec

from abc import abstractmethod

class BaseParametrization(Worker):

    def __init__(self, 
        name : str, 
        mesh : SurfaceMesh, 
        verbose : bool = False, 
        **kwargs
    ):
        super().__init__(name, verbose)
        self.mesh : SurfaceMesh = mesh
        self.save_on_corners : bool = kwargs.get("save_on_corners", True)
        self.uvs : ArrayAttribute = kwargs.get("uv_attr", None) # attribute on corners or vertices

        self._flat_mesh : SurfaceMesh = None

    @abstractmethod
    def run(self):
        pass

    @property
    def flat_mesh(self) -> SurfaceMesh:
        """
        A flat representation of the mesh where uv-coordinates are copied to xy.

        Returns:
            SurfaceMesh: the flat mesh
        """
        if self.uvs is None: return None
        if self._flat_mesh is None:
            # build the flat mesh : vertex coordinates are uv of original mesh
            self._flat_mesh = copy(self.mesh)
            for T in self.mesh.id_faces:
                for i,v in enumerate(self.mesh.faces[T]):
                    if self.save_on_corners:
                        self._flat_mesh.vertices[v] = Vec(self.uvs[3*T+i][0], self.uvs[3*T+i][1], 0.)
                    else:
                        self._flat_mesh.vertices[v] = Vec(self.uvs[v][0], self.uvs[v][1], 0.)
        return self._flat_mesh



from random import randint
from collections import deque

from ...mesh.datatypes import *
from ...attributes.misc_cells import cell_barycenter
from ...utils import keyify
from .base import SpanningTree, SpanningForest

class CellSpanningTree(SpanningTree):
    """
    A spanning tree defined over the connectivity of a volume mesh.

    Warning:
        This tree considers a unique starting point and stops when all reachable vertices are visited. 
        If the mesh is disconnected, the tree will be incomplete. In this case, use `CellSpanningForest` class instead.
    """

    @allowed_mesh_types(VolumeMesh)
    def __init__(self, mesh : VolumeMesh, starting_cell=None, forbidden_faces:set = None):
        super().__init__(mesh)
        if starting_cell is not None:
            self.root = starting_cell
        else:
            self.root = randint(0,len(self.mesh.cells)-1)
        
        if forbidden_faces is None:
            self.forbidden_faces = set()
        else:
            self.forbidden_faces : set = forbidden_faces
        self.parent = [None]*len(self.mesh.cells)
        self.children = [[] for v in self.mesh.id_cells]
        self.edges = []

    def compute(self):
        dist_to_root = [float("inf") for v in self.mesh.id_cells]
        seen = [False for _ in self.mesh.id_cells]
        queue = deque()

        def put_neighbours_in_queue(c):
            for F in self.mesh.connectivity.cell_to_face(c):
                if F in self.forbidden_faces: continue
                c2 = self.mesh.connectivity.other_face_side(c,F)
                if c2 is not None and not seen[c2]:
                    queue.append((c,c2))
        
        put_neighbours_in_queue(self.root)
        dist_to_root[self.root] = 0
        seen[self.root] = True

        while len(queue)>0:
            c,nc = queue.popleft()
            if seen[nc] : continue
            seen[nc] = True
            if dist_to_root[c] + 1 < dist_to_root[nc]:
                self.parent[nc] = c
                dist_to_root[nc] = dist_to_root[c] +1
            put_neighbours_in_queue(nc)

        # build children data from parent data
        for c in self.mesh.id_cells:
            p = self.parent[c]
            if p is not None:
                self.children[p].append(c)
                self.edges.append( keyify(p,c))
        super().compute() # sets the 'computed' flag

    def build_tree_as_polyline(self) -> PolyLine:
        """Builds the tree as a new polyline object. Useful for debug and visualization purposes

        Returns:
            PolyLine: the tree
        """
        output = PolyLine()
        if self.mesh.faces.has_attribute("barycenter"):
            bary = self.mesh.cells.attribute("barycenter")
        else:
            bary = cell_barycenter(self.mesh, persistent=False)
        for iC in self.mesh.id_cells:
            output.vertices.append(bary[iC])
            if self.parent[iC] is not None:
                output.edges.append([iC, self.parent[iC]])
        return output

class CellSpanningForest(SpanningForest):
    """A spanning forest that runs between cells. 
    Unlike a spanning tree, will create new roots and expand trees until all cells of the mesh have been visited
    """

    @allowed_mesh_types(VolumeMesh)
    def __init__(self, mesh : VolumeMesh):
        super().__init__(mesh)

    def compute(self) -> None : 
        visited = [False]*len(self.mesh.cells)
        for c in self.mesh.id_cells:
            if not visited[c]:
                # create a tree starting from c
                self.roots.append(c)
                tree_v = CellSpanningTree(self.mesh, c)()
                self.trees.append(tree_v)
                for (node, _) in tree_v.traverse():
                    visited[node] = True
        super().compute() # sets the 'computed' flag


"""Mouette

Maillages, Outils Et Traitement auTomatique de la géométriE
(Meshes, Tools and Geometry Processing)
"""

from . import config
from . import utils
from . import mesh
from . import attributes
from . import procedural
from . import operators
from . import geometry
from . import processing
from . import optimize
from . import sampling
from . import spatial
from . import splines

from .processing import framefield
from .processing import parametrization

from .geometry.vector import Vec
from .geometry import transform
from .utils import Logger
from .processing.worker import Worker
from .mesh.mesh_attributes import ArrayAttribute, Attribute



from .base import BaseParametrization

from ...mesh.datatypes import SurfaceMesh,PolyLine
from ...mesh.mesh_attributes import Attribute
from ...mesh.mesh import copy

from ... import operators
from ... import attributes
from .. import SingularityCutter
from ..trees import FaceSpanningTree
from ..connection  import SurfaceConnectionFaces

from ... import geometry as geom
from ...geometry import Vec

import numpy as np
import cmath
import scipy.sparse as sp

class ConformalConeParametrization(BaseParametrization):
    """
    Given a user-defined cone distribution on the surface, this algorithm cuts an input surface mesh into 
    a disk topology and parametrize it using conformal mapping

    References:
        - [1] _Conformal equivalence of triangle meshes_, Springborn B., Schröder P. and Pinkall U., ACM Transaction on Graphics, 2008
        
        - [2] _Boundary first flattening_, Sawhney R. and Crane K., ACM Transaction on Graphics, 2018
    """

    def __init__(self, mesh:SurfaceMesh, cones: Attribute, verbose:bool=False, **kwargs):
        """
        Args:
            mesh (SurfaceMesh): Input mesh
            cones (Attribute): float Attribute on vertices. Gives the target angle defects of vertices
            verbose (bool, optional): verbose mode. Defaults to False.
            
        Keyword Args:
            use_cotan (bool, optional): If True, uses cotangents in the laplacian matrix. Defaults to True. 
            debug (bool, optional) : debug mode. Generates additional output. Defaults to False.
        """
        super().__init__("ConeParam", mesh, verbose, **kwargs)
        self._use_cotan = kwargs.get("use_cotan", True)
        self._debug = kwargs.get("debug", False)
        self.save_on_corners = True
        assert cones.elemsize == 1
        assert cones.type == Attribute.Type.Float
        self._singus : Attribute = cones

        self._scale_factor : np.ndarray = None
        self._frames : np.ndarray = None
        self._conn : SurfaceConnectionFaces = None
        self._cutter : SingularityCutter
    
    def _check_singularity_validity(self):
        ### TODO
        return

    def run(self) :
        """
        Computes the parametrization
        """

        self.log("Perform cuts between cones")
        ## Define a cut mesh along the shortest path towards the boundary
        self._cutter = SingularityCutter(self.mesh, self._singus)
        self._cutter.run()

        self.log("Compute scale factor on vertices")
        ## Solve Yamabe problem to retrieve scale factor
        K = attributes.angle_defects(self.mesh, persistent=False).as_array(len(self.mesh.vertices))
        hK = self._singus.as_array(len(self.mesh.vertices))
        lap = operators.laplacian(self.mesh, cotan=self._use_cotan)
        self._scale_factor = sp.linalg.spsolve(lap, hK - K) # one scale factor per vertex
        self._scale_factor -= np.median(self._scale_factor) # log scale factor is defined up to a global constant (<=> the parametrization is defined up to a global scale)

        if self._debug:
            scale_fact_attr = self.mesh.vertices.create_attribute("scale", float, dense=True)
            for v in self.mesh.id_vertices:
                scale_fact_attr[v] = self._scale_factor[v]

        self.log("Integrate scale factor into local frames")
        cot = attributes.cotan_weights(self.mesh, dense=True, persistent=False)
        self._conn = SurfaceConnectionFaces(self.mesh)
        self._frames = np.zeros(len(self.mesh.faces), dtype=complex)

        ## Determine the root of the tree traversal. Should be a boundary face if possible
        tree_root = 0  
        edge_root = None # A boundary edge to be aligned vertically or horizontally
        if len(self.mesh.boundary_edges)>0:
            edge_root = self.mesh.boundary_edges[0]
            tree_root = [_T for _T in self.mesh.connectivity.edge_to_faces(*self.mesh.edges[edge_root]) if _T is not None][0]

        ## Perform tree traversal and integrate the scale factor
        tree = FaceSpanningTree(self.mesh, starting_face=tree_root, forbidden_edges=self._cutter.cut_edges)() # traverse the faces but avoid seams
        for face,parent in tree.traverse():
            if parent is None: # the tree root
                if edge_root is None:
                    self._frames[face] = complex(1.,0.)
                else:
                    A,B = self.mesh.edges[edge_root]
                    E = self.mesh.vertices[B] - self.mesh.vertices[A]
                    X,Y = self._conn.base(face)
                    c = complex(E.dot(X), E.dot(Y)) # edge in local basis coordinates
                    self._frames[face] = c/abs(c) # the edge gives the global frame reference
                continue

            A,B = self.mesh.connectivity.common_edge(face,parent)
            e = self.mesh.connectivity.edge_id(A,B)
            w = cot[e]* (self._scale_factor[B] - self._scale_factor[A])
            if self.mesh.connectivity.direct_face(A,B)==face: 
                w*=-1 # w is an oriented dual 1-form
            pt = self._conn.transport(face,parent)
            self._frames[face] = cmath.rect(1, w+pt) * self._frames[parent]


        self.log("Compute scaled edges in parameter space")
        scale_edges = np.zeros(2*len(self.cut_mesh.edges), dtype=float)
        cot = attributes.cotan_weights(self.cut_mesh, dense=True, persistent=False)
        for e,(A,B) in enumerate(self.cut_mesh.edges):
            rA,rB = self._cutter.ref_vertex[A], self._cutter.ref_vertex[B]
            E = self.mesh.vertices[rB] - self.mesh.vertices[rA]
            sce = np.exp( (self._scale_factor[rA] + self._scale_factor[rB])/2) # discrete scale factor of edge
            if self.cut_mesh.is_edge_on_border(A,B):
                T = [_T for _T in self.cut_mesh.connectivity.edge_to_faces(A,B) if _T is not None][0]
                X,Y = self._conn.base(T)
                ET = complex(X.dot(E), Y.dot(E))
                ET = ET / self._frames[T]
                scale_edges[2*e:2*e+2] = cot[e] * sce * Vec(ET.real, ET.imag)
            else:
                T1,T2 = self.cut_mesh.connectivity.edge_to_faces(A,B)
                X1,Y1 = self._conn.base(T1)
                ET1 = complex(X1.dot(E), Y1.dot(E))
                ET1 = ET1 / self._frames[T1]
                X2,Y2 = self._conn.base(T2)
                ET2 = complex(X2.dot(E), Y2.dot(E))
                ET2 = ET2 / self._frames[T2]
                mean = (ET1+ET2)/2
                scale_edges[2*e:2*e+2] = cot[e] * sce * Vec(mean.real, mean.imag)

        self.log("Solve for closest matching uv-coordinates")
        M = sp.kron(operators.cotan_edge_diagonal(self.cut_mesh, inverse=False) @ operators.vertex_to_edge_operator(self.cut_mesh, oriented=True).transpose(), sp.eye(2))
        UVs = sp.linalg.lsqr(M, scale_edges)[0]

        self.log("Write final result")
        ## Rescale UV coordinates to fit in [0,1]²
        xmin,xmax,ymin,ymax = float("inf"), -float("inf"), float("inf"), -float("inf")
        for v in self.cut_mesh.id_vertices:
            xmin = min(xmin, UVs[2*v])
            xmax = max(xmax, UVs[2*v])
            ymin = min(ymin, UVs[2*v+1])
            ymax = max(ymax, UVs[2*v+1])
        scale_x = xmax-xmin
        scale_y = ymax-ymin
        scale = min(scale_x, scale_y)

        ## Write final result in attribute
        self.uvs = self.mesh.face_corners.create_attribute("uv_coords", float, 2, dense=True)
        for c in self.mesh.id_corners:
            v = self.cut_mesh.face_corners[c]
            self.uvs[c] = Vec(UVs[2*v], UVs[2*v+1])/scale

    @property
    def frame_field(self) -> PolyLine:
        """Local frames of reference rendered as a polyline"""
        FFmesh = PolyLine()
        L = attributes.mean_edge_length(self.mesh)/3
        for id_face, face in enumerate(self.mesh.faces):
            basis,Y = self._conn.base(id_face)
            normal = geom.cross(basis,Y)
            pA,pB,pC = (self.mesh.vertices[_v] for _v in face)
            angle = cmath.phase(self._frames[id_face])
            bary = (pA+pB+pC)/3 # reference point for display
            r1,r2,r3,r4 = (geom.rotate_around_axis(basis, normal, angle + k*np.pi/2) for k in range(4))
            p1,p2,p3,p4 = (bary + abs(self._frames[id_face])*L*r for r in (r1,r2,r3,r4))
            FFmesh.vertices += [bary, p1, p2, p3, p4]
            FFmesh.edges += [(5*id_face, 5*id_face+k) for k in range(1,5)]
        return FFmesh  

    @property
    def cut_mesh(self) -> SurfaceMesh:
        """The mesh where seams have been disconnected and are now boundary edges

        Returns:
            SurfaceMesh: the mesh where seams have been disconnected
        """
        return self._cutter.output_mesh

    @property
    def cut_graph(self) -> PolyLine:
        """The seams returned as a polyline

        Returns:
            PolyLine: seams
        """
        return self._cutter.cut_graph

    @property
    def flat_mesh(self) -> SurfaceMesh:
        """
        A flat representation of the mesh where uv-coordinates are copied to xy.

        Returns:
            SurfaceMesh: the flat mesh
        """
        if self.uvs is None : return None
        if self._flat_mesh is None:
            # build the flat mesh : vertex coordinates are uv of original mesh
            self._flat_mesh = copy(self.cut_mesh)
            for T in self.cut_mesh.id_faces:
                for i,v in enumerate(self.cut_mesh.faces[T]):
                    self._flat_mesh.vertices[v] = Vec(self.uvs[3*T+i][0], self.uvs[3*T+i][1], 0.)
        return self._flat_mesh